# External
import os
# Internal
from dbgen.support.datatypes.table     import Table,Col,FK
from dbgen.support.datatypes.rule      import Rule,SBlock,SimpleUpdate,Plan,PureSQL,PBlock
from dbgen.support.datatypes.misc      import Arg
from dbgen.support.utils               import mkInsCmd,mkUpdateCmd
from dbgen.support.datatypes.sqltypes  import Varchar
from dbgen.inputs.extra                import surface

from dbgen.scripts.Pure.Atoms.get_bulk           import get_bulk
from dbgen.scripts.Pure.Atoms.get_miller         import get_miller
from dbgen.scripts.Pure.Atoms.surface_bulkstruct import surface_bulkstruct
from dbgen.scripts.Pure.Atoms.get_pure_struct    import get_pure_struct
from dbgen.scripts.Pure.Atoms.json_to_traj       import json_to_traj
from dbgen.scripts.Pure.Atoms.get_miller         import get_miller
from dbgen.scripts.Pure.Misc.pmg_to_ase          import pmg_to_ase
################################################################


species = Table('species'
    ,desc="""A combination of composition and structural information (no floats)"""
    ,cols=[Col('species_id',               pk   = True, auto = True)
          ,Col('name',          Varchar(), uniq = True)
          ,Col('nickname',      Varchar())
          ,Col('composition',   Varchar())])

bulkspecies = Table('bulkspecies'
    ,desc = 'Properties of species that are bulks'
    ,cols = [Col('species_id',    pk=True)
            ,Col('pure_struct_id',nn=True)]
    ,fks  = [FK('species_id',    'species')
            ,FK('pure_struct_id','pure_struct')])

molspecies = Table('molspecies'
    ,desc = 'Properties of species that are molecules. IN THE FUTURE, the SMILES string should be here because pointgroup + composition is underspecified'
    ,cols = [Col('species_id',pk = True)
            ,Col('pointgroup',nn = True)]
    ,fks  = [FK('species_id','species')
             ,FK('pure_struct_id','pure_struct')])

surfspecies = Table('surfspecies'
    ,desc = 'How should this be specified ??? And 2D materials?'
    ,cols = [Col('species_id',                  pk = True)
            ,Col('pure_struct_id',              nn = True)
            ,Col('facet',           Varchar(),  nn = True)]
    ,fks  = [FK('species_id','species')
            ,FK('pure_struct_id','pure_struct')])

species_composition = Table('species_composition'
    ,desc='Mapping table between species and element'
    ,cols = [Col('species_id',pk = True)
            ,Col('element_id',pk = True)
            ,Col('count',     nn = True)]
    ,fks  = [FK('species_id','species')
            ,FK('element_id','element')])

########################################################

pure_struct = Table('pure_struct'
    ,desc = 'Structure abstraction developed by Ankit Jain'
    ,cols = [Col('pure_struct_id',              pk = True,auto = True)
            ,Col('name',            Varchar(),  nn = True,uniq = True)
            ,Col('spacegroup')
            ,Col('free')
            ,Col('nickname',        Varchar())
            ])

tables = [species
         ,bulkspecies
         ,molspecies
         ,surfspecies
         ,species_composition
         ,pure_struct]

################################################################
################################################################
################################################################
nick_dict = {'AB_1_a_b_225'     : 'rocksalt'
            ,'AB_1_a_c_216'     : 'zincblende'
            ,'AB_1_a_b_221'     : 'cesium chloride'
            ,'A_1_a_225'        : 'fcc'
            ,'A_1_a_229'        : 'bcc'
            ,'A_2_c_194'        : 'hcp'
            ,'A_2_a_227'        : 'diamond'
            ,'AB3_1_a_d_221'    : 'anti-ReO3'
            ,'A2B3_8_ad_e_206'  : 'antibixbyite'
            # add rutile/perovskite
            }

pop_nickname = Rule('pop_nickname'
    ,plan=Plan(PureSQL(["""UPDATE pure_struct
                        SET   nickname = '{0}'
                        WHERE name     = '{1}'
                        AND pure_struct_id > 0
                        """.format(v,k) for k,v in nick_dict.items()])))

pop_compstr = Rule('species_compstr'
    ,plan=Plan(PureSQL([""" UPDATE species AS S SET S.composition=
                  (select  GROUP_CONCAT(CONCAT('[',C.element_id,',',C.count,']') ORDER BY C.element_id ASC)
                        FROM species_composition as C
                        WHERE C.species_id = S.species_id)"""])))

update_species_fk = Rule('species_fk'
    ,plan= Plan(PureSQL(["""UPDATE struct AS S SET S.species_id = (
                SELECT  P.species_id FROM species AS P
                LEFT JOIN bulkspecies B USING (species_id)
                LEFT JOIN molspecies M USING (species_id)
                WHERE S.composition_norm = P.composition
                 AND    (S.system_type = 'bulk'
                        AND B.pure_struct_id = S.pure_struct_id)
                        OR (S.system_type = 'molecule'
                        AND M.pointgroup = S.symmetry))
                WHERE S.struct_id > 0"""])))

pop_pure = Rule('pop_pure_struct'
    ,query = 'SELECT S.struct_id,S.raw FROM struct AS S WHERE pure_struct_id IS NULL'
    ,plan  = Plan([PBlock(get_bulk
                                        ,args=[Arg('json_to_traj')])
                                ,PBlock(json_to_traj
                                        ,args=[Arg('raw')])
                                ,PBlock(get_pure_struct
                                        ,args=[Arg('get_bulk')])
                ,SBlock(name = 'insert'
                             ,text = mkInsCmd('pure_struct',['name','spacegroup','free'],sqlite=False)
                             ,args = [Arg('get_pure_struct',i) for i in range(3)])
                        ,SBlock(name='query'
                            ,text="""SELECT pure_struct_id FROM pure_struct WHERE name=%s"""
                            ,args=[Arg('get_pure_struct',0)],deps=['insert'])
                        ,SBlock(name='insert_fk'
                                ,text=mkUpdateCmd('struct',['pure_struct_id'],['struct_id'])
                                ,args=[Arg('query',0),Arg('struct_id')])]))

underlying = Rule('underlying'
    ,query    = """SELECT U.struct_id,S.raw
                    FROM struct AS S
                    JOIN surface AS U USING (struct_id)
                    WHERE U.facet_h IS NULL"""
    ,plan = Plan([PBlock(json_to_traj,args = [Arg('raw')])
                 ,PBlock(surface_bulkstruct,args = [Arg('json_to_traj')])
                 ,PBlock(pmg_to_ase,args=[Arg('surface_bulkstruct')])
                 ,PBlock(get_bulk,args=[Arg('pmg_to_ase')])
                 ,PBlock(get_miller,args=[Arg('surface_bulkstruct')])
                 ,PBlock(get_pure_struct,args=[Arg('get_bulk')])
                 ,SBlock(name = 'insert'
                         ,text = mkInsCmd('pure_struct',['name','spacegroup','free'],sqlite=False)
                         ,args = [Arg('get_pure_struct',i) for i in range(3)])
                 ,SBlock(name='query'
                    ,text="""SELECT pure_struct_id FROM pure_struct WHERE name=%s"""
                    ,args=[Arg('get_pure_struct',0)],deps=['insert'])
                 ,SBlock(name='insert_fk'
                        ,text=mkUpdateCmd('surface',['pure_struct_id'],['struct_id'])
                        ,args=[Arg('query',0),Arg('struct_id')])
                 ,SBlock(name='update_miller'
                      ,text = mkUpdateCmd('surface',['facet_h','facet_k','facet_l'],['struct_id'])
                      ,args = [Arg('get_miller',i) for i in range(3)]
                             +[Arg('struct_id')])]))

rules = [pop_pure
             ,underlying
             ,pop_nickname
             ,pop_compstr
             ,update_species_fk]



from typing     import Tuple,List
from os         import environ
from subprocess import check_output

def populate_storage()->Tuple[List[str],List[str],List[str]]:
    """
    Reads DFT_LOGFILES to find a structured textfile with information
    about what directories should be considered 'roots'
    """
    if environ['USER'] == 'ksb':
       print('warning, using smallroots')
       path = '/scratch/users/ksb/share/reference/smallroots.txt'
    else:
        path = '/scratch/users/ksb/share/reference/roots.txt'
    user = environ['SHERLOCK2_USERNAME']
    sher = '%s@login.sherlock.stanford.edu'%user
    data = check_output('ssh %s cat %s'%(sher,path),shell=True).decode('utf-8')
    rootdata  = [l.split()
                    for l in data.split('\n')
                    if '#' not in l and l.strip()!='']
    paths,codes,labels = zip(*rootdata)
    return list(paths),list(codes),list(labels)



# External Modules
from typing import List
import os

# Internal Modules
from dbgen.support.datatypes.table     import Table,Col,FK
from dbgen.support.datatypes.table     import Table,Col,FK
from dbgen.support.datatypes.table     import Table,Col,FK,View
from dbgen.support.datatypes.misc      import Arg,noIndex
from dbgen.inputs.basics               import struct,traj,atom
from dbgen.core.lists                  import flatten,merge_dicts
from dbgen.support.datatypes.sqltypes  import Varchar,Decimal
from dbgen.support.datatypes.rule      import (Rule,PureSQL,SimpleUpdate,Plan
                                              ,PBlock,Const,SBlock,SimpleFunc
                                              ,SimpleInsert,SimplePipe,Block
                                              ,Unpack)

from dbgen.scripts.Pure.Load.get_dipol_gpaw         import get_dipol_gpaw
from dbgen.scripts.Pure.Load.get_dipol_qe           import get_dipol_qe
from dbgen.scripts.Pure.Load.get_dipol_vasp         import get_dipol_vasp
from dbgen.scripts.Pure.Load.get_gamma_gpaw         import get_gamma_gpaw
from dbgen.scripts.Pure.Load.get_gamma_qe           import get_gamma_qe
from dbgen.scripts.Pure.Load.get_gamma_vasp         import get_gamma_vasp
from dbgen.scripts.Pure.Load.get_econv_gpaw         import get_econv_gpaw
from dbgen.scripts.Pure.Load.get_econv_qe           import get_econv_qe
from dbgen.scripts.Pure.Load.get_econv_vasp         import get_econv_vasp
from dbgen.scripts.Pure.Load.get_sigma_gpaw         import get_sigma_gpaw
from dbgen.scripts.Pure.Load.get_sigma_qe           import get_sigma_qe
from dbgen.scripts.Pure.Load.get_sigma_vasp         import get_sigma_vasp
from dbgen.scripts.Pure.Load.get_mixing_gpaw        import get_mixing_gpaw
from dbgen.scripts.Pure.Load.get_mixing_qe          import get_mixing_qe
from dbgen.scripts.Pure.Load.get_mixing_vasp        import get_mixing_vasp
from dbgen.scripts.Pure.Load.get_nbands_gpaw        import get_nbands_gpaw
from dbgen.scripts.Pure.Load.get_nbands_qe          import get_nbands_qe
from dbgen.scripts.Pure.Load.get_nbands_vasp        import get_nbands_vasp
from dbgen.scripts.Pure.Load.get_nmix_gpaw          import get_nmix_gpaw
from dbgen.scripts.Pure.Load.get_nmix_qe            import get_nmix_qe
from dbgen.scripts.Pure.Load.get_nmix_vasp          import get_nmix_vasp
from dbgen.scripts.Pure.Load.get_diag_gpaw          import get_diag_gpaw
from dbgen.scripts.Pure.Load.get_diag_qe            import get_diag_qe
from dbgen.scripts.Pure.Load.get_diag_vasp          import get_diag_vasp
from dbgen.scripts.Pure.Load.get_spinpol_gpaw       import get_spinpol_gpaw
from dbgen.scripts.Pure.Load.get_spinpol_qe         import get_spinpol_qe
from dbgen.scripts.Pure.Load.get_spinpol_vasp       import get_spinpol_vasp
from dbgen.scripts.Pure.Load.get_dipole_gpaw        import get_dipole_gpaw
from dbgen.scripts.Pure.Load.get_dipole_qe          import get_dipole_qe
from dbgen.scripts.Pure.Load.get_dipole_vasp        import get_dipole_vasp
from dbgen.scripts.Pure.Load.get_maxstep_gpaw       import get_maxstep_gpaw
from dbgen.scripts.Pure.Load.get_maxstep_qe         import get_maxstep_qe
from dbgen.scripts.Pure.Load.get_maxstep_vasp       import get_maxstep_vasp

from dbgen.scripts.Pure.Load.get_kpts_gpaw          import get_kpts_gpaw
from dbgen.scripts.Pure.Load.get_kpts_qe            import get_kpts_qe
from dbgen.scripts.Pure.Load.get_kpts_vasp          import get_kpts_vasp
from dbgen.scripts.Pure.Misc.dftcode_dependent      import dftcode_dependent
from dbgen.scripts.Pure.Misc.identity               import identity
from dbgen.scripts.Pure.Misc.get                    import get
from dbgen.scripts.Pure.Misc.flatten                import flatten
from dbgen.scripts.Pure.Misc.map_                   import map_
from dbgen.scripts.Pure.Misc.pair                   import pair
from dbgen.scripts.Pure.Misc.triple                 import triple
from dbgen.scripts.Pure.Atoms.get_voronoi           import get_voronoi
from dbgen.scripts.Pure.Atoms.get_pointgroup        import get_pointgroup
from dbgen.scripts.Pure.Atoms.get_spacegroup        import get_spacegroup
from dbgen.scripts.Pure.Atoms.json_to_traj          import json_to_traj
from dbgen.scripts.Pure.Atoms.get_vacuum            import get_vacuum
from dbgen.scripts.Pure.Load.load_comp              import load_comp
from dbgen.scripts.Pure.Load.get_vasp_key           import get_vasp_key
from dbgen.scripts.Pure.Load.get_dw                 import get_dw
from dbgen.scripts.Pure.Load.get_cell_dofree        import get_cell_dofree
from dbgen.scripts.Pure.Load.get_cell_factor        import get_cell_factor
from dbgen.scripts.Pure.Load.get_mixtype            import get_mixtype
from dbgen.scripts.Pure.Load.get_kptden             import get_kptden
from dbgen.scripts.Pure.Graph.mk_graph              import mk_graph
from dbgen.scripts.Pure.Graph.layers                import layers
from dbgen.scripts.Pure.Graph.geometric             import geometric
from dbgen.scripts.Pure.Graph.graph_to_json         import graph_to_json
from dbgen.scripts.Pure.Graph.json_to_graph         import json_to_graph

################################################################
################################################################
################################################################
# CONSTANTS
###########

dfts         = ['gpaw','qe','vasp']
other_cols = ['dipol','gamma','econv','sigma','mixing','nbands'
             ,'nmix','diag','spinpol','dipole','maxstep','kpts']
cell_cols = ['a0','a1','a2','b0','b1','b2','c0','c1','c2']



get_fns = [get_dipol_gpaw,get_dipol_qe,get_dipol_vasp,get_gamma_gpaw
          ,get_gamma_qe,get_gamma_vasp,get_econv_gpaw,get_econv_qe,get_econv_vasp
          ,get_sigma_gpaw,get_sigma_qe,get_sigma_vasp,get_mixing_gpaw
          ,get_mixing_qe,get_mixing_vasp,get_nbands_gpaw,get_nbands_qe,get_nbands_vasp
          ,get_nmix_gpaw,get_nmix_qe,get_nmix_vasp,get_diag_gpaw,get_diag_qe,get_diag_vasp
          ,get_spinpol_gpaw,get_spinpol_qe,get_spinpol_vasp,get_dipole_gpaw
          ,get_dipole_qe,get_dipole_vasp,get_maxstep_gpaw,get_maxstep_qe,get_maxstep_vasp
          ,get_kpts_gpaw,get_kpts_qe,get_kpts_vasp]

################################################################
################################################################
################################################################

################################################################
################################################################
################################################################
struct_composition = Table('struct_composition'
        ,desc = 'Chemical Composition of a structure'
        ,cols = [Col('struct_id',             pk=True)
                ,Col('element_id',            pk=True)
                ,Col('has_fixed',             nn=True)
                ,Col('count',                 nn=True)
                ,Col('count_fixed',           nn=True)
                ,Col('count_norm',            nn=True)
                ,Col('frac',        Decimal(),nn=True)
                ,Col('frac_fixed',  Decimal(),nn=True)
                ,Col('fixed_norm',            nn=True)
                ]
        ,fks = [FK('struct_id', 'struct')
               ,FK('element_id','element')])

################################################################
surface = Table('surface'
        ,desc = 'Subclass of struct'
        ,cols = [Col('struct_id', pk=True)
                ,Col('vacuum',     Decimal())
                ,Col('pure_struct_id') # underlying structure
                ,Col('facet_h')
                ,Col('facet_k')
                ,Col('facet_l')
                ,Col('pure_struct_id_catalog')
                ,Col('facet_h_catalog')
                ,Col('facet_k_catalog')
                ,Col('facet_l_catalog')
                ]
        ,fks=[FK('struct_id','struct')
             ,FK('pure_struct_id','pure_struct')
             ,FK('pure_struct_id_catalog','pure_struct','pure_struct_id')])

bulk = Table('bulk'
        ,desc = 'Subclass of struct'
        ,cols = [Col('struct_id',    pk=True)
                ,Col('pure_struct_id_catalog')]
        ,fks=[FK('struct_id','struct')])

molecule = Table('molecule'
        ,desc = 'Subclass of struct'
        ,cols = [Col('struct_id',        pk=True)
                ,Col('pointgroup',  Varchar())]
        ,fks=[FK('struct_id','struct')])

########################################################################

################################################################################


similar_struct= Table('similar_struct'
    ,desc = "Store a precomputation for identifying related jobs"
    ,cols = [Col('struct_id',pk=True)
            ,Col('struct_id2',pk=True)]
    ,fks  = [FK('struct_id','struct')
            ,FK('struct_id2','struct','struct_id')])

############################################################################

calc_other = Table('calc_other'
        ,desc = """Less important DFT Calculator Parameters"""
        ,cols = [Col('calc_other_id',pk=True,auto=True)
                ,Col("kx")
                ,Col("ky")
                ,Col("kz")
                ,Col("fmax",            Decimal())
                ,Col("econv",           Decimal(10,7))
                ,Col("dw")
                ,Col("sigma",           Decimal(10,7))
                ,Col("nbands")
                ,Col("mixing",          Decimal())
                ,Col("nmix")
                ,Col("xtol",            Decimal(10,7))
                ,Col("strain",          Decimal())
                ,Col("gga",             Varchar())
                ,Col("luse_vdw")
                ,Col("zab_vdw",         Decimal())
                ,Col("nelmdl")
                ,Col("gamma")
                ,Col("dipol",           Varchar())
                ,Col("algo",            Varchar())
                ,Col("ibrion")
                ,Col("prec",            Varchar())
                ,Col("ionic_steps")
                ,Col("lreal",           Varchar())
                ,Col("lvhar")
                ,Col("diag",            Varchar())
                ,Col("spinpol")
                ,Col("dipole")
                ,Col("maxstep")
                ,Col("delta",           Decimal())
                ,Col("mixingtype",      Varchar())
                ,Col("bonded_inds",     Varchar())
                ,Col("energy_cut_off",  Decimal())
                ,Col("step_size",       Decimal())
                ,Col("spring",          Decimal())
                ,Col('cell_dofree',     Varchar())
                ,Col('cell_factor',     Decimal())
                ,Col("kpts",            Varchar())])


#############################################################################
vib_job = Table('vib_job'
        ,desc = 'Jobs that compute vibrational modes for a struct'
        ,cols = [Col('job_id',pk=True)
                ,Col('calc_id')
                ,Col('struct_id')]
        ,fks  = [FK('job_id',    'alljob')
                ,FK('calc_id',    'calc')
                ,FK('struct_id','struct')])

dos_job = Table('dos_job'
        ,desc = 'Jobs that compute density of states for a struct'
        ,cols = [Col('job_id',pk=True)
                ,Col('calc_id')
                ,Col('struct_id')]
        ,fks  = [FK('job_id',    'alljob')
                ,FK('calc_id',    'calc')
                ,FK('struct_id','struct')])

neb_job = Table('neb_job'
        ,desc = 'Jobs that compute NEB between two structures'
        ,cols = [Col('job_id',pk=True)
                ,Col('calc_id')
                ,Col('init_struct_id')
                ,Col('final_struct_id')]
        ,fks  = [FK('job_id',    'alljob')
                ,FK('calc_id',    'calc')
                ,FK('init_struct_id','struct','struct_id')
                ,FK('final_struct_id','struct','struct_id')])

bulkmod_job = Table('bulkmod_job'
        ,desc = 'Jobs that compute bulk modulus for a struct'
        ,cols = [Col('job_id',        pk=True)
                ,Col('calc_id')
                ,Col('struct_id')
                ,Col('bulkmod',    Decimal())]
        ,fks  = [FK('job_id',    'alljob')
                ,FK('calc_id',  'calc')
                ,FK('struct_id','struct')])

tables = [struct_composition
         ,surface
         ,bulk
         ,molecule
         ,calc_other
         ,similar_struct
         ,dos_job
         ,vib_job
         ,neb_job
         ,bulkmod_job]
##############################################################################
##############################################################################
##############################################################################


geograph = Rule('geograph'
    ,query = """SELECT S.struct_id,S.raw from struct AS S WHERE S.geo_graph IS NULL"""
    ,plan  = Plan(blocks = [PBlock(geometric,args=[Arg('json_to_traj')])
                           ,PBlock(json_to_traj,args=[Arg('raw')])
                           ,PBlock(graph_to_json
                                  ,'geo_graph'
                                  ,args=[Arg('mk_graph')])
                           ,PBlock(mk_graph
                                  ,args = noIndex(['json_to_traj'
                                                  ,'include_frac'
                                                  ,'group_cut'
                                                  ,'min_bo'
                                                  ,'geometric']))
                           ,SimpleUpdate(struct,['geo_graph'])]
                    ,consts = {'include_frac': Const(0.8)
                                 ,'group_cut': Const(0.3)
                                 ,'min_bo'   : Const(0.03)}))

elemental = Rule('elemental'
    ,plan=Plan(PureSQL(["""UPDATE struct AS S SET S.elemental =
                             S.n_atoms=1 AND S.system_type = 'molecule'
                             WHERE S.struct_id >0"""])))
fmax = Rule('fmax'
    ,plan=Plan(PureSQL(["""UPDATE traj AS T SET T.fmax =

                (SELECT
                    MAX(SQRT(TA.fx * TA.fx + TA.fy * TA.fy + TA.fz * TA.fz)) AS fmax
                FROM
                    traj AS T2
                JOIN
                    trajatom AS TA USING (job_id , traj_id , struct_id)
                JOIN
                    atom AS A USING (struct_id , atom_id)
                WHERE
                    A.constrained = 0
                GROUP BY job_id,traj_id)


                WHERE
                T2.job_id=T.job_id AND T2.traj_id=T.traj_id
                AND T.job_id >0 AND T.struct_id > 0"""])))

voronoi = Rule('voronoi'
    ,query      = """SELECT DISTINCT S.struct_id,S.raw
                     FROM struct AS S JOIN atom AS A USING (struct_id)
                     WHERE A.v3 IS NULL"""
    ,plan     = Plan(blocks=[PBlock(get_voronoi
                                          ,args = [Arg('json_to_traj')])
                                 ,PBlock(json_to_traj
                                          ,args = [Arg('raw')])
         ,SBlock(name='update_voronoi'
                                ,text="""UPDATE atom SET v3=%s,v4=%s,v5=%s,v6=%s
                                                        ,v7=%s,v8=%s,v9=%s,v10=%s
                                        WHERE struct_id = %s AND atom_id = %s"""
                                ,args = [Arg('get_voronoi',i) for i in range(1,9)]
                                        +[Arg('struct_id'),Arg('get_voronoi',0)])]))

populate_subjobs = Rule(name = 'subjobs'
        ,plan = Plan(PureSQL([
        """ INSERT INTO bulkmod_job (job_id)
                SELECT J.job_id FROM alljob AS J
                    WHERE J.job_type = 'bulkmod'
                    AND J.job_id NOT IN (SELECT B.job_id from bulkmod_job as B)
                    AND NOT J.deleted"""
        ,""" INSERT INTO chargemol_job (job_id)
                SELECT J.job_id FROM alljob AS J
                    WHERE J.job_type = 'chargemol'
                    AND J.job_id NOT IN (SELECT C.job_id from chargemol_job as C)
                    AND NOT J.deleted"""
        ,""" INSERT INTO vib_job (job_id)
                SELECT J.job_id FROM alljob AS J
                    WHERE J.job_type = 'vib'
                    AND J.job_id NOT IN (SELECT V.job_id from vib_job as V)
                    AND NOT J.deleted"""
        ,""" INSERT INTO neb_job (job_id)
                SELECT J.job_id FROM alljob AS J
                    WHERE J.job_type = 'neb'
                    AND J.job_id NOT IN (SELECT N.job_id from neb_job as N)
                    AND NOT J.deleted"""
        ,""" INSERT INTO dos_job (job_id)
                SELECT J.job_id FROM alljob AS J
                    WHERE J.job_type = 'dos'
                    AND J.job_id NOT IN (SELECT D.job_id from dos_job as D)
                    AND NOT J.deleted"""])))

populate_substructs = Rule(name = 'substructs'
    ,plan = Plan(PureSQL([
    """INSERT INTO molecule (struct_id)
        SELECT S.struct_id from struct AS S
        WHERE S.system_type = 'molecule'
        AND S.struct_id NOT IN (SELECT molecule.struct_id from molecule)"""
    ,"""INSERT INTO bulk (struct_id)
        SELECT S.struct_id from struct AS S
        WHERE S.system_type = 'bulk'
        AND S.struct_id NOT IN (SELECT bulk.struct_id from bulk)"""
    ,"""INSERT INTO surface (struct_id)
        SELECT S.struct_id from struct AS S
                        WHERE S.system_type = 'surface'
                        AND S.struct_id NOT IN (SELECT surface.struct_id from surface)"""])))


comp_cols = ['element_id','has_fixed'
            ,'count','count_fixed','count_norm'
            ,'frac','frac_fixed','fixed_norm']

# if any rows with a particular job have been added, we assume all have been added
populate_struct_composition = Rule(name='composition'
    ,query = """SELECT S.struct_id,S.raw
                FROM  struct AS S
                WHERE S.struct_id NOT IN (SELECT C.struct_id FROM struct_composition AS C)"""
    ,plan    = Plan(SimpleFunc(load_comp,['raw'],comp_cols)
                   +SimpleInsert(struct_composition,['struct_id']+comp_cols))
    )

comp_strs = Rule(name='composition_strs'
    ,plan = Plan(PureSQL(["""
        UPDATE struct AS S SET S.composition_norm =
        (select  GROUP_CONCAT(CONCAT('[',C.element_id,',',C.count_norm,']') ORDER BY C.element_id ASC)
                FROM struct_composition as C
                WHERE C.struct_id = S.struct_id)

            ,S.composition=
              (select  GROUP_CONCAT(CONCAT('[',C.element_id,',',C.count,']') ORDER BY C.element_id ASC)
                    FROM struct_composition as C
                    WHERE C.struct_id = S.struct_id)

            ,S.metal_comp=
                (select  GROUP_CONCAT(CONCAT('[',C.element_id,',',C.count,']') ORDER BY C.element_id ASC)
                      FROM struct_composition as C
                      WHERE C.struct_id = S.struct_id
                      AND C.element_id NOT IN (1,2,6,7,8,9,10,16,17,18))
        WHERE S.struct_id > 0"""
            ])))


update_pointgroup = Rule(name='update_pointgroup'
    ,query = """SELECT M.struct_id,S.raw
                FROM molecule AS M
                    JOIN struct AS S USING (struct_id)
                WHERE M.pointgroup IS NULL"""
    ,plan   = Plan(SimplePipe([json_to_traj,get_pointgroup],['raw'],['pointgroup'])
                +[SimpleUpdate(molecule,['pointgroup'])]))

update_spacegroup = Rule(name='update_spacegroup'
    ,query = """SELECT S.struct_id,S.raw
                    FROM struct AS S
                WHERE S.spacegroup IS NULL"""
    ,plan   = Plan(SimplePipe([json_to_traj,get_spacegroup],['raw'],['spacegroup'])
                  +[SimpleUpdate(struct,['spacegroup'])]))

#######################

# Calc other cols only relevant for VASP jobs
calc_other_pairs = {'prec'      : 'str'
                   ,'luse_vdw'  : 'bool'
                   ,'gga'       : 'bool'
                   ,'lvhar'     : 'bool'
                   ,'zab_vdw'   : 'float'
                   ,'algo'      : 'str'
                   ,'nelmdl'    : 'int'
                   ,'ibrion'    : 'int'
                   ,'lreal'     : 'str'
                   }

# Calc other cols only relevant for VASP
vasp_key_tempfuncs = [PBlock(get_vasp_key
         ,name = x
         ,args = noIndex(['dftcode','log','key '+x,'cast '+y]))
                            for x,y in calc_other_pairs.items()]  # type: List[Block]

# Calc other cols relevant for all three codes
dftdep_tempfuncs = [PBlock(dftcode_dependent
                            ,name = x
                            ,args = noIndex(['dftcode'
                                            ,'log'
                                            ,'get_%s_gpaw'%x
                                            ,'get_%s_qe'%x
                                            ,'get_%s_vasp'%x
                                            ,'identity'
                                            ,'identity'])) for x in other_cols if x not in ['nbands','maxstep','kpts','gamma']] # type: List[Block]
# Calc other cols relevant for all three codes that depend on files other than log
dftdep_dict_tempfuncs = [PBlock(dftcode_dependent
                                 ,name = x
                                 ,args = noIndex(['dftcode'
                                                  ,'files'
                                                  ,'get_%s_gpaw'%x
                                                  ,'get_%s_qe'%x
                                                  ,'get_%s_vasp'%x
                                                  ,'identity'
                                                  ,'identity'])) for x in  ['maxstep','kpts','gamma']] # type: List[Block]
# Calc other cols that can 'only' be scraped from params.json
get_pairs = {'fmax'              :'float'
            ,'strain'            :'float'
            ,'energy_cut_off'    :'float'
            ,'bonded_inds'       :'str'
            ,'step_size'         :'float'
            ,'spring'            :'float'
            ,'delta'             :'float'
            }

get_tempfuncs = [PBlock(get
                         ,name = x
                         ,args = noIndex(['paramdict','key '+x,'cast '+y])) for x,y in get_pairs.items()] # type: List[Block]

qe_tempfuncs = [PBlock(x,x.__name__[4:],args = noIndex(['pwinp','dftcode']))
                    for x in [get_cell_factor,get_cell_dofree]] # type: List[Block]

# All calc_other column names
calc_other_cols = (list(calc_other_pairs.keys())
                 + list (get_pairs.keys())
                 + [o for o in other_cols if o != 'kpts']
                 + ['kx','ky','kz','cell_dofree','cell_factor'])

calc_other_args = noIndex(calc_other_cols)
double_co_args  = [val for val in calc_other_args for _ in (0, 1)]


other = [PBlock(identity,'kx',args=[Arg('kpts',0)])
             ,PBlock(identity,'ky',args=[Arg('kpts',1)])
             ,PBlock(identity,'kz',args=[Arg('kpts',2)])
             ,PBlock(get_dw,'dw',args=noIndex(['dftcode','log']))
             ,PBlock(get_mixtype,'mixtype',args=noIndex(['dftcode','log']))
             ,PBlock(pair,args=noIndex(['log','paramdict']))
             ,PBlock(triple,'files',args=noIndex(['log','pwinp','kptcar']))
             ,PBlock(dftcode_dependent
                       ,name = 'nbands'
                      ,args = noIndex(['dftcode'
                                      ,'pair' # THIS IS WHAT IS DIFFERENT
                                      ,'get_nbands_gpaw'
                                      ,'get_nbands_qe'
                                      ,'get_nbands_vasp'
                                      ,'identity'
                                      ,'identity']))]   # type: List[Block]
calc_other_p = other + dftdep_tempfuncs + dftdep_dict_tempfuncs + vasp_key_tempfuncs  + get_tempfuncs + qe_tempfuncs
calc_other_s = [SBlock(name = 'insert_calc'
                      ,text = """INSERT  into calc_other ({0})
                                 SELECT {1} FROM DUAL
                                WHERE NOT EXISTS (SELECT (1) FROM calc_other WHERE {2})
                                """.format(','.join(calc_other_cols)
                                          ,','.join(['%s']*len(calc_other_cols))
                                          ,' AND '.join(['({0}=%s OR ({0} IS NULL AND %s IS NULL ))'.format(x) for x in calc_other_cols]))
                      ,args = calc_other_args+double_co_args)
    ,SBlock(name = 'get_calc'
           ,text = """SELECT calc_other_id FROM calc_other WHERE {0}
                    """.format(' AND '.join(['({0}=%s OR ({0} IS NULL AND %s IS NULL ))'.format(x) for x in calc_other_cols]))
           ,args = double_co_args, deps = ['insert_calc'])
    , SBlock(name = 'update'
           ,text = "UPDATE relax_job SET calc_other_id = %s WHERE job_id=%s"
           ,args = [Arg('get_calc',0),Arg('job_id')])]    # type: List[Block]
# Main rule
populate_calc_other = Rule(name='calc_other'
    ,query    = """SELECT R.job_id,J.log,J.pwinp,J.kptcar,J.paramdict,C.dftcode
                     FROM relax_job AS R
                     JOIN alljob AS J USING (job_id)
                     JOIN calc AS C  USING (calc_id)
                     WHERE R.calc_other_id IS NULL"""
    ,plan = Plan(consts = merge_dicts([{'key ' +x:Const(x) for x in calc_other_pairs.keys()} # type: ignore
                                             ,{'cast '+y:Const(y) for y in calc_other_pairs.values()}
                                             ,{'key ' +x:Const(x) for x in get_pairs.keys()}
                                             ,{'cast '+y:Const(y) for y in get_pairs.values()}
                                             ,{'identity' : Const(identity)}
                                             ]+[{x.__name__ : Const(x)} for x in get_fns])

                 ,blocks = calc_other_p + calc_other_s))

kden_cols = ['kptden_x','kptden_y','kptden_z']

kptden = Rule(name='kptden'
    ,query = """SELECT T.job_id,T.traj_id,O.kx,O.ky,O.kz%s
                FROM traj AS T
                    JOIN struct    AS S USING (struct_id)
                    JOIN cell        AS C USING (cell_id)
                    JOIN relax_job AS J USING (job_id)
                    JOIN calc_other AS O USING (calc_other_id)
                WHERE T.kptden_x IS null
            """%(''.join([',C.%s'%c for c in cell_cols]))
    ,plan = Plan(SimpleFunc(get_kptden
                          ,['kx','ky','kz']+cell_cols
                          ,kden_cols)
                +[SimpleUpdate(traj,kden_cols)]))


populate_struct_maps = Rule(name='struct_maps'
    ,plan=Plan(PureSQL(["""
    INSERT INTO chargemol_map (job_id,charge_id)

    SELECT
        F.job_id,CJ.job_id
    FROM
        chargemol_job AS CJ
            JOIN struct AS CS USING (struct_id)
            JOIN cell AS CC USING (cell_id)
            JOIN similar_struct AS SS USING (struct_id) -- pairs of structures with same sym, chemical symbols, and constrained atoms (all non-float data)
            JOIN finaltraj AS F ON F.struct_id = SS.struct_id2
            JOIN struct    AS S ON F.struct_id = S.struct_id
            JOIN cell      AS C ON C.cell_id = S.cell_id
    WHERE
        ABS(CC.a0 - C.a0) < 0.01
            AND ABS(CC.a1 - C.a1) < 0.01
            AND ABS(CC.a2 - C.a2) < 0.01
            AND ABS(CC.b0 - C.b0) < 0.01
            AND ABS(CC.b1 - C.b1) < 0.01
            AND ABS(CC.b2 - C.b2) < 0.01
            AND ABS(CC.c0 - C.c0) < 0.01
            AND ABS(CC.c1 - C.c1) < 0.01
            AND ABS(CC.c2 - C.c2) < 0.01
            AND NOT EXISTS( SELECT (1)
                FROM  atom AS A JOIN atom AS CA
                WHERE   A.struct_id = S.struct_id
                    AND CA.struct_id = CS.struct_id
                    AND A.atom_id = CA.atom_id
                    AND (  ABS(A.x - CA.x) > 0.01
                        OR ABS(A.y - CA.y) > 0.01
                        OR ABS(A.z - CA.z) > 0.01))

        ON DUPLICATE KEY UPDATE job_id=F.job_id"""]))) # could replace 'chargemol' with 'vib' and reuse code


struct_strs = Rule('struct_strs'
    ,plan=Plan(PureSQL(["""
         UPDATE struct AS S SET S.str_symbols =
            (SELECT GROUP_CONCAT(A.element_id)
                FROM atom as A WHERE A.struct_id = S.struct_id)
  ,S.str_constraints =
    (SELECT GROUP_CONCAT(A.constrained)
        FROM atom as A WHERE A.struct_id = S.struct_id)

,S.symmetry = COALESCE((SELECT S.spacegroup
                FROM bulk AS B
                WHERE B.struct_id=S.struct_id)
            ,(SELECT M.pointgroup
                FROM molecule AS M
                WHERE M.struct_id=S.struct_id)
            ,'1')

    WHERE S.struct_id > 0            """])))

pop_similar_struct = Rule('pop_similar_struct'
    ,plan = Plan(PureSQL(["""INSERT INTO similar_struct  (struct_id,struct_id2)
                        SELECT S.struct_id,T.struct_id
                        FROM struct AS S JOIN struct AS T
                        WHERE S.str_symbols = T.str_symbols
                        AND S.str_constraints=T.str_constraints
                        AND S.symmetry = T.symmetry
                        ON DUPLICATE KEY UPDATE similar_struct.struct_id=similar_struct.struct_id"""])))

update_vacuum = Rule('update_vacuum'
    ,query = """SELECT U.struct_id,S.raw
                    FROM struct AS S JOIN surface AS U USING (struct_id)
                WHERE U.vacuum IS NULL"""
    ,plan   = Plan([PBlock(get_vacuum,'vacuum',[Arg('json_to_traj')])
                   ,PBlock(json_to_traj,args = [Arg('raw')])
                   ,SimpleUpdate(surface,['vacuum'])]))

update_layer = Rule('update_layer'
    ,query = """SELECT U.struct_id,S.geo_graph
                FROM struct AS S
                JOIN surface AS U USING (struct_id)
                JOIN atom AS A USING (struct_id)
                GROUP BY U.struct_id
            HAVING SUM(A.layer is null)!=0"""
    ,plan=Plan([PBlock(json_to_graph,args=[Arg('geo_graph')])
               ,PBlock(layers,args=[Arg('json_to_graph')])
               ,SimpleUpdate(atom,['layer'])]
               +Unpack('layers',['atom_id','layer'])))


pop_kpts = Rule('pop_kpts'
    ,plan=Plan(PureSQL(["""UPDATE calc_other AS C
                        SET  kpts = CONCAT('[', C.kx, ',', C.ky, ',', C.kz, ']')
                        WHERE C.calc_other_id > 0"""])))
rules = [geograph
             ,elemental
             ,voronoi
             ,populate_subjobs
             ,populate_substructs
             ,populate_struct_composition
             ,comp_strs
             ,update_pointgroup
             ,update_spacegroup
             ,update_vacuum
             ,populate_calc_other
             ,kptden
             ,populate_struct_maps
             ,struct_strs
             ,pop_similar_struct
             ,update_layer
             ,pop_kpts]



#External Modules
from typing import Callable,List,Optional,Union,Dict
#Internal Modules
from dbgen.support.datatypes.object         import View
from dbgen.support.datatypes.insertupdate   import InsertUpdate

from dbgen.support.datatypes.rule       import Rule,Plan,Const,PBlock,Block
from dbgen.support.datatypes.misc       import Arg
from dbgen.core.lists                   import flatten


Actions = List[Union[PBlock,List[PBlock]
              ,Block,List[Block]
              ,InsertUpdate,List[InsertUpdate]]]
Action = Union[PBlock,Block,InsertUpdate]
################################################################################

class Relation(object):
    """
    Relate attributes of objects via triples of:
    -select:  Extracting info from the current state of the world
    -inserts: putting information back into the DB
    """
    def __init__(self
                ,name    : str
                ,view    : Optional[View] = None
                ,actions : Actions   = []
                ,consts  : Dict[str,Const] = {}
                ) -> None:
        self.name    = name
        self.view    = view
        self.uni     = view.uni if view else {}
        self.consts  = consts

        # Take mixed singles + lists and make one list
        #---------------------------------------------
        acts = [] # type: List[Action]
        for a in actions:
            if isinstance(a,list):
                acts.extend(a)
            else:
                acts.append(a)
        self.actions = acts

    def mk_rule(self)->Rule:
        """

        """

        # Scenario: we want to insert into a table that is not mentioned in FROM
        # but is related by a FK (so we need identity "job_id = job_job_id")
        extra_fks = []
        if self.view:
            for o in self.view.allobj:
                if o not in self.view._from.allobj:
                    # This Object was not mentioned in FROM clause
                    # Search for any FKs linking this Object to any Object in FROM
                    for o2 in self.view._from.allobj:
                        try:
                            fk = self.view.get_fk(o,o2)
                            extra_fks.append(fk)
                        except ValueError:
                            pass

        # Need to include tempfuncs that capture equality across FKs
        if self.view:
            join_tfs = [j.fk.rename() for j in self.view._from.joins]
            xtra_tfs = [fk.rename() for fk in extra_fks]
            self.actions.extend(flatten(join_tfs)+flatten(xtra_tfs))

        blocks = [] # type: List[Block]
        for a in self.actions:
            if isinstance(a,Block):
                blocks.append(a)
            elif isinstance(a,InsertUpdate):
                blocks.extend(a.mk_block())

        return Rule(name     = self.name
                   ,query    = str(self.view) if self.view else None
                   ,plan     = Plan(blocks = blocks
                                   ,consts = self.consts))



# Internal modules
from dbgen.support.datatypes.model      import Model
from dbgen.support.datatypes.func       import Func
from dbgen.support.datatypes.object     import Object
from dbgen.new_inputs.relations        import relations

###############################################################################
if __name__=='__main__':
    from dbgen.main import parser
    args = parser.parse_args()
    m = Model(Object.universe,relations)
    m.run(reset   = args.reset
         ,catalog = args.catalog
         ,add     = args.add
         ,xclude  = args.xclude
         ,only    = args.only)



from dbgen.support.datatypes.object     import Object,NOTNULL,FK,UNIQUE,DEFAULT,view
from dbgen.support.datatypes.sqltypes   import SQLType,Int,Varchar,Text,Decimal
#############################################################################
class Atoms(Object):
    """

    """
    element_id = Varchar()
    x          = Decimal()
    y          = Decimal()
    z          = Decimal()
    ox         = Decimal()
    fx         = Decimal()
    fy         = Decimal()
    fz         = Decimal()
    magmom     = Decimal()
    charge     = Decimal()
    volume     = Decimal()
    occupancy  = Decimal()

class Authors(Object):
    """

    """
    first_name = Varchar()
    last_name  = Varchar()

class Calculations(Object):
    """

    """
    configuration   = Varchar()
    label           = Varchar()
    path            = Varchar()
    settings        = Varchar()
    energy          = Decimal()
    energy_pa       = Decimal()
    magmom          = Decimal()
    magmom_pa       = Decimal()
    band_gap        = Decimal()
    attempt         = Int()
    nsteps          = Int()
    converged       = Int()
    runtime         = Int()
    natoms          = Int()
    irreducible_kpoints  = Int()


class Calculations_element_set(Object):
    """

    """


class Calculations_hubbard_set(Object):
    """

    """


class Calculations_metadata(Object):
    """

    """


class Calculations_potential_set(Object):
    """

    """


class Compositions(Object):
    """

    """


class Compositions_element_set(Object):
    """

    """


class Dos(Object):
    """

    """


class Elements(Object):
    """

    """


class Entries(Object):
    """

    """


class Entries_element_set(Object):
    """

    """


class Entries_metadata(Object):
    """

    """


class Entries_species_set(Object):
    """

    """


class Expt_formation_energies(Object):
    """

    """


class Fits(Object):
    """

    """


class Fits_dft(Object):
    """

    """


class Fits_elements(Object):
    """

    """


class Fits_experiments(Object):
    """

    """


class Formation_energies(Object):
    """

    """


class Hubbard_corrections(Object):
    """

    """


class Hubbards(Object):
    """

    """


class Journals(Object):
    """

    """


class Metadata(Object):
    """

    """


class Operations(Object):
    """

    """


class Prototypes(Object):
    """

    """


class Publications(Object):
    """

    """


class Publications_author_set(Object):
    """

    """


class Reference_energies(Object):
    """

    """


class Rotations(Object):
    """

    """


class Sites(Object):
    """

    """


class Spacegroups(Object):
    """

    """


class Spacegroups_centering_vectors(Object):
    """

    """


class Spacegroups_operations(Object):
    """

    """


class Species(Object):
    """

    """


class Structures(Object):
    """

    """


class Structures_element_set(Object):
    """

    """


class Structures_metadata(Object):
    """

    """


class Structures_species_set(Object):
    """

    """


class Translations(Object):
    """

    """


class Vasp_potentials(Object):
    """

    """


class Wyckoffsites(Object):
    """

    """



# External modules
from ase        import Atoms    # type: ignore
from ase.io     import read     # type: ignore
from os.path    import join,exists,getsize
from os         import environ,mkdir,listdir
from subprocess import call
from shutil     import rmtree
from string     import ascii_uppercase,digits
import tempfile
# Internal modules
from dbgen.scripts.Pure.Atoms.traj_to_json import traj_to_json
################################################################################
def anytraj(root : str) -> str:
    """
    ASE IO read function - takes any traj it can find
    """
    user = environ['SHERLOCK2_USERNAME']
    temp_dir = tempfile.mkdtemp()

    call(['scp','-C','-r','-q','%s@login.sherlock.stanford.edu:%s/*.traj'%(user,root)
         ,temp_dir])
    trajs = [join(temp_dir,t) for t in listdir(temp_dir) if getsize(join(temp_dir,t)) > 100]

    if len(trajs) == 0:
        raise ValueError('Get Traj could not find any traj in '+root)
    else:
        atoms = read(trajs[0])
        rmtree(temp_dir)
        return traj_to_json(atoms)

if __name__=='__main__':
    import sys
    print(anytraj(sys.argv[1]))



# External Modules
from os import environ
# Internal Modules
from dbgen.inputs.species              import nick_dict
from dbgen.core.lists                  import merge_dicts
from dbgen.support.datatypes.rule      import (Rule,PureSQL,SimpleUpdate,Plan
                                              ,PBlock,Const,SBlock,SimpleFunc
                                              ,SimplePipe,Block
                                              ,Unpack)
from dbgen.support.datatypes.misc      import Arg,noIndex
from dbgen.support.utils               import mkInsCmd,mkUpdateCmd
from dbgen.inputs.basics               import alljob

from dbgen.scripts.Pure.Misc.get                import get
from dbgen.scripts.Pure.Misc.get_               import get_
from dbgen.scripts.Pure.Load.get_catalog_ads    import get_catalog_ads
from dbgen.scripts.Pure.Load.get_catalog_facet  import get_catalog_facet
"""
Anything that seems Catalog-specific (absolutely useless for jobs generated
elsewhere) should go here
"""
################################################################################

# Mapping from informal structure names given by user to specific pure_struct's
struct_dict = merge_dicts([{v:k,v+'_unit':k,v+'-unit':k} for k,v in nick_dict.items()])
# Additional cases
struct_dict['2DMaterial'] = 'N/A'
struct_dict['hexagonal'] = struct_dict['hcp']

pop_bulk_struct_catalog = Rule('pop_bulk_struct_catalog'
    ,query = """SELECT B.struct_id,J.paramdict
                FROM bulk AS B
                JOIN struct AS S USING (struct_id)
                JOIN traj USING (struct_id)
                JOIN alljob AS J USING (job_id)
                WHERE J.paramdict != ''
                AND S.pure_struct_id_catalog IS NULL
                AND NOT J.deleted"""
    ,plan = Plan([PBlock(get
                          ,name='nickname'
                          ,args=noIndex(['paramdict'
                                        ,'structure_key'
                                        ,'str']))
                 ,PBlock(get_
                          ,'pure_struct_name'
                          ,args=noIndex(['struct_dict'
                                        ,'nickname'
                                        ,'unknown']))
                ,SBlock(text = mkInsCmd('pure_struct',['name'])
                       ,name = 'insert_pure_struct'
                       ,args = [Arg('pure_struct_name')])
                ,SBlock(text = """SELECT pure_struct_id FROM pure_struct WHERE name=%s"""
                       ,name = 'query_pure_struct'
                       ,args = [Arg('pure_struct_name')],deps=['insert_pure_struct'])
                ,SBlock(text = mkUpdateCmd('struct',['pure_struct_id_catalog'],['struct_id'])
                       ,name = 'update_fk'
                       ,args = noIndex(['query_pure_struct','struct_id']))]

            ,{'struct_dict'   : Const(struct_dict)
                                       ,'structure_key' : Const('structure')
                                       ,'str'           : Const('str')
                                       ,'unknown'       : Const('unknown')}))

pop_surf_struct_catalog = Rule('pop_surf_struct_catalog'
    ,query="""SELECT S.struct_id,J.paramdict
                FROM surface AS S
                JOIN traj USING (struct_id)
                JOIN alljob AS J USING (job_id)
                WHERE J.paramdict != ''
                AND S.facet_l_catalog IS NULL
                AND NOT J.deleted """
    ,plan=Plan([PBlock(get
                      ,name='nickname'
                      ,args=noIndex(['paramdict'
                                    ,'structure_key'
                                    ,'str']))

             ,PBlock(get_catalog_facet,args=[Arg('paramdict')])

             ,PBlock(get_
                      ,'pure_struct_name'
                      ,args=noIndex(['struct_dict'
                                    ,'nickname'
                                    ,'unknown']))

            ,SBlock(name = 'insert_pure_struct'
                   ,text = mkInsCmd('pure_struct',['name'])
                   ,args=[Arg('pure_struct_name')])

           ,SBlock(name = 'query_pure_struct'
                  ,text = """SELECT pure_struct_id FROM pure_struct WHERE name=%s"""
                  ,args = [Arg('pure_struct_name')],deps=['insert_pure_struct'])

           ,SBlock(name ='update_fk'
                  ,text =mkUpdateCmd('surface'
                             ,['pure_struct_id_catalog'
                              ,'facet_h_catalog'
                              ,'facet_k_catalog'
                              ,'facet_l_catalog']
                             ,['struct_id'])
                 ,args =[Arg('query_pure_struct')
                        ,Arg('get_catalog_facet',0)
                        ,Arg('get_catalog_facet',1)
                        ,Arg('get_catalog_facet',2)
                        ,Arg('struct_id')])]

    ,{'struct_dict'     : Const(struct_dict)
                             ,'structure_key'   : Const('structure')
                             ,'str'             : Const('str')
                             ,'unknown'         : Const('unknown')}))

pop_ads_catalog = Rule('pop_ads_catalog'
    ,query = """SELECT J.job_id,S.struct_id,J.paramdict
                FROM surface AS S
                JOIN traj USING (struct_id)
                JOIN alljob AS J USING (job_id)
                WHERE J.paramdict like '%%adsorbates%%'
                AND J.ads_catalog IS NULL
                AND NOT J.deleted"""
    ,plan=Plan([PBlock(get
                        ,name='ads_json'
                        ,args=noIndex(['paramdict'
                                        ,'ads'
                                        ,'str']))
               ,PBlock(get_catalog_ads,args=[Arg('ads_json')])
               ,SBlock(text=mkUpdateCmd('atom'
                                  ,['struct_adsorbate_id_catalog']
                                  ,['struct_id','atom_id'])
                      ,name='insert'
                    ,args=[Arg('get_catalog_ads',0)
                          ,Arg('struct_id')
                          ,Arg('get_catalog_ads',1)])
                ,SBlock(name='insert2'
                    ,text=mkUpdateCmd('alljob',['ads_catalog'],['job_id'])
                    ,args=noIndex(['ads_json','job_id']))]
            ,{'ads': Const('adsorbates')
             ,'str': Const('str')}))


pop_struct_catalog = Rule('pop_struct_catalog'
    ,query = """SELECT J.job_id,J.paramdict
                FROM alljob AS J
                WHERE J.paramdict like '%%structure%%'
                AND J.structure_catalog IS NULL
                AND NOT J.deleted """
    ,plan=Plan(SimpleFunc(get,['paramdict','struct','str'],['structure_catalog'])
            +[SimpleUpdate(alljob,['structure_catalog'])]
    ,{'struct': Const('structure')
     ,'str'   : Const('str')}))

rules = [pop_bulk_struct_catalog
             ,pop_surf_struct_catalog
             ,pop_ads_catalog
             ,pop_struct_catalog]



from dbgen.support.datatypes.table     import Table,Col,FK
from dbgen.support.datatypes.table     import Table,Col,FK,View
from dbgen.support.datatypes.misc      import Arg,noIndex
from dbgen.support.datatypes.rule      import Rule,Plan,PureSQL,PBlock,SBlock
from dbgen.support.datatypes.sqltypes  import Varchar,Decimal

from dbgen.support.utils               import mkInsCmd
from dbgen.core.lists                  import flatten

from dbgen.scripts.Pure.Graph.json_to_graph  import json_to_graph
from dbgen.scripts.Pure.Graph.identify_ads   import identify_ads
##########################################################################################
##########################################################################################
##########################################################################################
# Constants
###########
cell_cols = ['a0','a1','a2','b0','b1','b2','c0','c1','c2']

ads = {'H':{1:1},'N':{7:1},'O':{8:1},'C':{6:1},'NH':{1:1,7:1},'OH':{8:1,1:1}
      ,'CH':{1:1,6:1},'NH2':{1:2,7:1},'H2O':{1:2,8:1},'CH2':{1:2,6:1}
      ,'NH3':{1:3,7:1},'CH3':{1:3,6:1},'N2':{7:2},'NNH':{1:1,7:2},'NNH2':{1:2,7:2}
      ,'NNH3':{1:3,7:2},'H2':{1:2},'C2':{6:2},'N2':{7:2},'O2':{8:2}
      ,'OOH':{1:1,8:2},'CHO':{1:1,6:1,8:1},'COOH':{1:1,6:1,8:2}
      ,'CO':{6:1,8:1},'OCCO':{6:2,8:2},'OCCHO':{1:1,6:2,8:2}
      }

##########################################################################################
##########################################################################################
##########################################################################################

adsorbate = Table('adsorbate'
    ,desc = 'Species that can adsorb onto a surface'
    ,cols = [Col('adsorbate_id',            pk = True, auto = True)
            ,Col('name',        Varchar(),  nn = True, uniq = True)
            ,Col('composition', Varchar())])

adsorbate_composition = Table('adsorbate_composition'
    ,'Components of an adsorbate'
    ,cols = [Col('adsorbate_id',  pk = True)
            ,Col('element_id',    pk = True)
            ,Col('number',        nn = True)])

struct_adsorbate = Table('struct_adsorbate'
    ,'An adsorbate considered on a particular surface'
    ,cols = [Col('struct_id',                   pk=True)
            ,Col('adsorbate_id',                pk=True)
            ,Col('struct_adsorbate_id',         pk=True)
            ,Col('site',                Varchar())]
    ,fks = [FK('struct_id',   'struct')
           ,FK('adsorbate_id','adsorbate')])

tables = [adsorbate
         ,adsorbate_composition
         ,struct_adsorbate]

##########################################################################################
##########################################################################################
##########################################################################################

ads_eng_view = View("ads_eng","""
SELECT R.ref_scheme_id                   -- SPECIFY *1* OF THESE
      ,A.adsorbate_id                    -- JOIN ON THIS
      ,J.calc_id                         -- JOIN ON THIS
      ,R.element_id                      -- SUM (number*avg_energy) WHEN GROUPING BY THIS
      ,R.component_id                    -- AVG energy IN SUBSELECT WHILE GROUPING BY THIS + Element&Component
      ,A.name                            -- human readable info
      ,E.symbol AS element               -- human readable info
      ,CE.symbol AS component            -- human readable info
      ,RS.struct_id                      -- join with struct for info to be used in constraints (e.g. kpt related things depending on bulk/molecule/surface)
      ,F.job_id                          -- join with JOB to be used in constraints (e.g. timestamp constraint to get repeatable results)
      ,F.energy/C.count AS energy_norm   -- contains fundamental information we want (averaged for each component element and summed to get adsorbate reference energy)
      ,J.calc_other_id                   -- join with calc other to filter down possible contributions to adsorbate energy
      ,AC.number                         -- how many of ELEMENT are in the adsorbate
      ,R.coef                            -- how much of each COMPONENT refjob goes into the calculation of the energy per ELEMENT atom
    FROM adsorbate_composition AS AC
    JOIN ref_scheme_reqs AS R ON AC.element_id = R.element_id
    JOIN adsorbate AS A USING (adsorbate_id)
    JOIN ref_scheme_refstruct AS RS ON RS.element_id=R.component_id AND RS.ref_scheme_id = R.ref_scheme_id
    JOIN finaltraj AS F ON F.struct_id = RS.struct_id
    JOIN relax_job AS J ON J.job_id = F.job_id
    JOIN element AS E ON E.element_id=R.component_id
    JOIN element AS CE ON CE.element_id=R.component_id
    JOIN struct_composition AS C ON C.element_id=RS.element_id AND C.struct_id = F.struct_id""")
views = [ads_eng_view]
##########################################################################################
##########################################################################################
##########################################################################################


adscompstr  = "((SELECT adsorbate_id FROM adsorbate WHERE name='%s'),%d,%d)"
adscompstrs = flatten([[adscompstr%(k,k2,v2) for k2,v2 in v.items()] for k,v in ads.items()])
populate_adsorbate = Rule('populate_adsorbate'
    , plan=Plan(PureSQL(["""INSERT INTO adsorbate (name) VALUES %s
                            ON DUPLICATE KEY UPDATE name=name"""%','.join(["('%s')"%x for x in ads.keys()])
                       ,"""INSERT INTO adsorbate_composition (adsorbate_id,element_id,number) VALUES %s
                           ON DUPLICATE KEY UPDATE adsorbate_id=adsorbate_id"""%','.join(adscompstrs)])))


#####
struct_ads_cols = ['struct_id','adsorbate_id','struct_adsorbate_id','site']
pop_struct_ads = Rule('pop_struct_ads'
    ,query = """SELECT S.struct_id,S.geo_graph
                FROM struct AS S JOIN surface USING (struct_id)
                WHERE EXISTS (SELECT A.struct_adsorbate_id IS NULL
                                AND A.element_id IN (1,6,7,8)
                              FROM atom AS A
                              WHERE A.struct_id = S.struct_id)"""

    ,plan= Plan([PBlock(json_to_graph,args=[Arg('geo_graph')])
                ,PBlock(identify_ads, 'ads', [Arg('json_to_graph')])
                ,SBlock(name = 'adsorbate_id'
                       ,text = 'SELECT A.adsorbate_id FROM adsorbate AS A WHERE A.name=%s'
                       ,args = [Arg('ads',4)])
                ,SBlock(name='adsorbate_id_many'
                       ,text ='SELECT A.adsorbate_id FROM adsorbate AS A WHERE A.name=%s'
                       ,args = [Arg('ads',2)])
                ,SBlock(name = 'insert_struct_ads'
                       ,text = mkInsCmd('struct_adsorbate',struct_ads_cols,sqlite=False)
                       ,args = [Arg('struct_id')
                              ,Arg('adsorbate_id')
                              ,Arg('ads',3)
                              ,Arg('ads',5)])
                ,SBlock(name = 'update_atom_fk'
                       ,text = """UPDATE atom SET struct_id=%s,adsorbate_id=%s,struct_adsorbate_id=%s
                                 WHERE struct_id=%s AND atom_id=%s"""
                       ,args = [Arg('struct_id') # needed because it's a composite foreign key?
                              ,Arg('adsorbate_id_many')
                              ,Arg('ads',1)
                              ,Arg('struct_id')
                              ,Arg('ads',0)])]))

pop_ads_compstr = Rule('pop_ads_compstr'
    ,plan=Plan(PureSQL(["""
            UPDATE adsorbate AS A SET A.composition=
              (select  GROUP_CONCAT(CONCAT('[',C.element_id,',',C.number,']') ORDER BY C.element_id ASC)
                    FROM adsorbate_composition as C
                    WHERE C.adsorbate_id = A.adsorbate_id)"""])))

rules = [populate_adsorbate
             ,pop_struct_ads
             ,pop_ads_compstr]



from dbgen.support.datatypes.rule      import Rule,Plan,SBlock,Const,SimpleFunc
from dbgen.support.datatypes.misc      import Arg,noIndex
from dbgen.support.datatypes.sqltypes  import Varchar,Decimal

from dbgen.support.utils               import mkInsCmd
from dbgen.support.datatypes.table      import Table,Col,FK
from dbgen.scripts.IO.parse_keld import parse_keld


##########################################################################################
keld_table = Table('keld_solids'
    ,desc='Data from Keld L. with experimental properties for bulk materials'
    ,cols = [Col('id',                            pk=True,auto=True)
            ,Col('name',             Varchar(),       nn=True)
            ,Col('job_name',         Varchar(),       nn=True)
            ,Col('lattice_parameter',Decimal())
            ,Col('cohesive_energy',  Decimal())
            ,Col('bulkmod',          Decimal())
            ,Col('structure',        Varchar())
            ,Col('magmom',           Decimal())])

tables = [keld_table]

##########################################################################################
##########################################################################################
##########################################################################################

keld_cols2 = ['name_short','pure_struct_id','spacegroups'
             ,'name_mid','element_id','count'
             ,'name_long','property','value']

s_d_cols = ['struct_dataset_id','species_id','property','value']

pop_keld = Rule('keld'
    ,plan = Plan(SimpleFunc(parse_keld,[],keld_cols2)

        + [SBlock(name = 'insert_species'
              ,text = mkInsCmd('species',['name','nickname'])
              ,args = noIndex(['name_short','name_short']))

        ,SBlock(name = 'insert_pure_struct'
              ,text = mkInsCmd('pure_struct',['name','spacegroup'])
              ,args = noIndex(['pure_struct_id','spacegroups']))

        ,SBlock(name = 'insert_bulk_species'
              ,text = mkInsCmd('bulkspecies',['species_id','pure_struct_id'],sqlite=False)
              ,args = [Arg('query_species_short')
                      ,Arg('query_pure_struct')])

        ,SBlock(name = 'insert_dataset'
              ,text = """INSERT INTO struct_dataset (name) VALUES (%s)
                      ON DUPLICATE KEY UPDATE struct_dataset_id=struct_dataset_id"""
              ,args = [Arg('k')])

        ,SBlock(name = 'query_species_short'
              ,text = 'SELECT species_id FROM species WHERE name=%s'
              ,args = [Arg('name_short')],deps=['insert_species'])

        ,SBlock(name = 'query_pure_struct'
              ,text = 'SELECT pure_struct_id FROM pure_struct WHERE name=%s'
              ,args = [Arg('pure_struct_id')],deps=['insert_pure_struct'])

        ,SBlock(name = 'query_species_long'
              ,text = 'SELECT species_id FROM species WHERE name=%s'
              ,args = [Arg('name_long')], deps = ['insert_species'])

        ,SBlock(name = 'query_dataset'
              ,text = """SELECT  struct_dataset_id FROM struct_dataset
                        WHERE name = %s"""
              ,args = [Arg('k')],deps=['insert_dataset'])

        ,SBlock(name = 'query_species_mid'
              ,text = 'SELECT species_id FROM species WHERE name=%s'
              ,args = [Arg('name_mid')],deps=['insert_species'])

        ,SBlock(name = 'insert_species_comp'
              ,text = mkInsCmd('species_composition',['species_id','element_id','count'],sqlite=False)
              ,args = [Arg('query_species_mid')
                      ,Arg('element_id')
                      ,Arg('count')])
                      
        ,SBlock(name = 'insert_dataset_elem'
              ,text = mkInsCmd('struct_dataset_element',s_d_cols,sqlite=False)
              ,args = [Arg('query_dataset')
                      ,Arg('query_species_long')
                      ,Arg('property'),Arg('value')])]
        ,{'k':Const('keld_solids')}))

rules = [pop_keld]



from typing     import List,Tuple
from os         import environ
from os.path    import join,exists
from subprocess import check_output,DEVNULL

str9 = Tuple[str,str,str,str,str,str,str,str,str]
################################################################################
def get_catalog_logfiles(rootpath : str
                    ) -> '''Tuple[List[str],List[str],List[str]
                                 ,List[str],List[str],List[str]
                                 ,List[str],List[str],List[str]]''':
    """
    Searches for DFT calculation folders. Returns unzipped (stordir,logfile,code) triples

    """
    # Auxillary Functions
    #--------------------
    def process(dir:str,log:str,pw:str,pot:str,pos:str,out:str,kpt:str,param:str)->str9:
        if out:
            code    = 'vasp'
            logfile = dir+'/OUTCAR'
            log     = out
        elif pw:
            code    ='quantumespresso'
            logfile = dir+'/log'
        else:
            code    = 'gpaw'
            logfile = dir+'/log'
        return dir,logfile,code,log,pw,pot,pos,kpt,param

    # Initialize Variables
    #---------------------
    user   = environ['SHERLOCK2_USERNAME']
    login  = '%s@login.sherlock.stanford.edu'%user
    delim1 = '***???***'
    delim2 = '~~~***~~~'
    d1     = ';echo "%s";done;'%delim1
    d2     = 'echo "%s";cat $f/'%delim2
    loop   = r'for f in %s/*;'%rootpath
    dirs,logfiles,codes,logs,pws,pots,pos,kpts,params = [['']]*9

    # Main Program
    #-------------
    cmd  = '{2} do echo $f;{1}log;{1}pw.inp;{1}POTCAR;{1}POSCAR;{1}OUTCAR;{1}KPOINTS;{1}params.json{0}'.format(d1,d2,loop,delim2)
    output =  check_output(['ssh', login,cmd],stderr=DEVNULL).decode('utf-8')
    fdata = [process(*map(str.strip,o.split(delim2))) for o in output.split(delim1)[1:-1]]

    dirs,logfiles,codes,logs,pws,pots,pos,kpts,params = zip(*fdata)

    return list(dirs),list(logfiles),list(codes),list(logs),list(pws),list(pots),list(pos),list(kpts),list(params)

if __name__ == '__main__':
    import sys
    print(get_catalog_logfiles(sys.argv[1]))



# External Modules
from typing          import Optional,List,Dict,Any,Callable,Set
from abc             import abstractmethod
from time            import time
from pprint          import pformat
from multiprocessing import Pool,cpu_count
from traceback       import format_exc
from decimal         import Decimal
from tqdm            import tqdm                       # type: ignore
from sqlparse        import format as sql_format       # type: ignore
from networkx        import DiGraph                    # type: ignore

# Internal Modules
from dbgen.core.lists   import flatten,concat_map
from dbgen.core.parsing import btw
from dbgen.core.misc    import identity
from dbgen.core.numeric import safe_div

from dbgen.support.sqlparsing         import parse_dependencies
from dbgen.support.datatypes.table    import Table
from dbgen.support.datatypes.func     import Func
from dbgen.support.datatypes.misc     import ExternalError,Arg,noIndex
from dbgen.support.utils              import (sqlexecute,sqlselect,mkInsCmd,sqlite_execute
                                             ,namespaceCmd,func_to_ind,sqlite_select
                                             ,handle_to_ind,ConnectInfo,select_dict
                                             ,sqlexecutemany,addQs,topsort_with_dict)
#########################################################################################
################################################################################
class Const(object):
    """
    Embed a constant value of any type into the namespace for a given rule.
    """
    def __init__(self,val:Any)->None:
        self.val = val
    def __str__(self)->str:
        return pformat(self.__dict__)
    def __eq__(self,other : object)->bool:
        return self.__dict__==other.__dict__
    def get_datatype(self)->str:
        return str(self.val.__class__)

    def add_const(self
                 ,sqlite_pth : str
                 ,trans_id   : int
                 ,const_id   : int
                 ,const_name : str
                 ) -> None:
        """
        Add to meta.db
        """
        datatype = self.get_datatype()
        cols     = ['rule_id','const_id','name','datatype','val']
        q        = mkInsCmd('constants',cols,sqlite=True)
        sqlite_execute(sqlite_pth,q,[trans_id,const_id,const_name,datatype,str(self.val)])
################################################################################

class Block(object):
    """
    Either a PBlock or an SBlock
    """
    # Common Properties of all Blocks
    #-------------------------------
    name = ''
    args = []    # type: List[Arg]
    deps = set() # type: Set[str]
    @abstractmethod
    def __init__(self)->None:
        raise NotImplementedError

    @abstractmethod
    def apply(self
             ,curr_dict : Dict[str,Any]
             ,cxn       : Optional[ConnectInfo] = None
             ) -> Any:
        raise NotImplementedError

################################################################################
class SBlock(Block):
    """
    SQL executing block
    """
    # Class constants
    #----------------
    valid_types = (int,str,float,list,bytes,Decimal,type(None))
    type_err    = "Arg (%s) BAD DATATYPE %s IN NAMESPACE "
    broad_err   = "Can't broadcast: maxlen = %d, len a = %d (%s)"
    sql_err     = 'SBlock needs a SQL connection'

    def __init__(self
                ,text : str
                ,name : Optional[str] = None
                ,args : List[Arg]     = []
                ,deps : List[str]     = []
                ) -> None:
        assert isinstance(text,str), 'Not str: '+str(text)
        if name is None:
            name = text
        self.text = text
        self.name = name
        self.args = args
        self.deps = set(deps)

        self.is_select = text[:6].lower()=='select'

    def apply(self
             ,curr_dict : Dict[str,Any]
             ,cxn       : Optional[ConnectInfo] = None
             ) -> Any:
        """
        Creates a function which takes in a namespace dictionary and executes its
        block script (possibly using the namespace for inputs). Its result will
        get added to the namespace.
        """
        assert cxn,self.sql_err
        args   = [arg.arg_get(curr_dict) for arg in self.args]
        maxlen = 1

        for a in args:
            assert isinstance(a,self.valid_types), self.type_err%(a,a.__class__)
            if isinstance(a,list):
                if maxlen != 1: # variable has been set
                    try:
                        assert(len(a) in [1,maxlen]), self.broad_err%(maxlen,len(a),str(a)) # preconditions for broadcasting
                    except:
                        import pdb;pdb.set_trace()
                else:
                    maxlen = len(a)

        def process_arg(x:Any)->list:
            if isinstance(x,list) and len(x)!=maxlen:
                return maxlen*x # broadcast
            elif not isinstance(x,list):
                return  maxlen * [x]
            else:
                return x

        # now all args should be lists of the same length
        broadcasted = [process_arg(x) for x in args]

        binds = list(zip(*broadcasted))

        if self.is_select:
            # we're probably selecting to get a PK, take the first result no matter what
            def app(x:tuple)->List[tuple]:
                    return sqlselect(cxn,self.text,x)[0] # type: ignore

            def mbBox(x:tuple)->Any:
                if len(x)==1: return x[0]
                else: return x
            uniqbinds = list(set(binds))
            try:
                output = {b:mbBox(app(b)) for b in uniqbinds} # type: ignore
            except IndexError:
                    print('failed to find a result after query failed:\n'+self.text)
                    print('binds are : '+pformat(list(zip(self.args,uniqbinds[0]))))
                    import pdb;pdb.set_trace()
                    return []

            return [output.get(x) for x in binds] # type: ignore
        else:
            if len(self.args)==0:
                sqlexecute(cxn,self.text)
            else:
                sqlexecutemany(cxn,self.text,binds)
            return None

# SBlock Shortcuts
#----------------
def PureSQL(strs : List[str])->List[Block]:
    """Execute a series of SQL statements"""
    return [SBlock(text=s,name="pure sql #%d"%i) for i,s in enumerate(strs)]

def SimpleInsert(tab:Table,cols:List[str]) -> List[Block]:
    """
    Assume that the insert columns have corresponding variables in the namespace
    """
    assert cols, 'Cannot insert with no columns'
    args     = noIndex(cols)
    qmarks   = ','.join(['%s']*len(cols))
    fmt_args = [tab.name,','.join(cols),qmarks,cols[0]]
    ins = SBlock(text = """INSERT INTO {0} ({1}) VALUES ({2})
                            ON DUPLICATE KEY UPDATE {3} = {3}""".format(*fmt_args)
                ,name = 'insert_'+tab.name,args = args) # type: Block

    pk_cols   = [c.name for c in tab.cols if c.pk == True]
    fmt_args2 = [','.join(pk_cols),tab.name,addQs(cols,' AND ')]
    q   = SBlock(text="""SELECT {0} FROM {1} WHERE {2}""".format(*fmt_args2)
                ,name = 'select_'+tab.name,args=args,deps=[ins.name]) # type: Block
    u   = Unpack('select_'+tab.name,[tab.name+'_'+c for c in pk_cols])
    return [ins]#,q]+u

def SimpleUpdate(tab:Table,cols:List[str]) -> Block:
    """
    Assume that the insert columns AND primary keys have corresponding variables
    in the namespace
    """
    pk_cols = [c.name for c in tab.cols if c.pk == True]
    args    = noIndex(cols+pk_cols)
    set     = addQs(cols,',')
    where   = addQs(pk_cols,' AND ')

    return SBlock(text='UPDATE %s SET %s WHERE %s'%(tab.name,set,where)
                 ,name='update_'+tab.name,args=args)

def SimpleUpsert(ins_tab   : Table
                ,ins_cols  : List[str]
                ,fk_tab    : Table
                ) -> List[Block]:
    """
    Let's say we compute an entity (A) associated with another entity (B).
    - We want to insert the A instance into the A table (if it doesn't exist).
    - We want to record the PK of this A instance an store it as a FK in B table.

    ins_tab   - 'A' table described above
    cols       - properties with which to insert an A instance
    fk_tab    - name of the 'B' table described above

    We assume that:
        - there exists exactly one foreign key from B to A
        - whatever was inserted will uniquely identify an A object (e.g. 'raw'
            for structure, possibly not necessarily the PKs of A)
        - namespace has appropriate column names

    WARNING THIS WILL BE A PROBLEM IF THERE'S A NAMESPACE COLLISION IN KEY NAMES

    SAFER WOULD BE TO MAKE THE 'key' FOR COLUMNS TO BE '<table>.<column>'
    """

    insert_args   = noIndex(ins_cols)
    ins_select    = SimpleInsert(ins_tab,ins_cols)

    ins_pks = [c.name for c in ins_tab.cols if c.pk]
    where  = addQs(ins_cols,' AND ')
    fmt_args1 = [','.join(ins_pks),ins_tab.name,where]
    get_id_block = SBlock('SELECT {0} FROM {1} WHERE {2}'.format(*fmt_args1)
                          ,'get_id_'+ins_tab.name,insert_args,[ins_select[0].name])

    fk         = fk_tab.get_fk(ins_tab)
    fk_pks     = [c.name for c in fk_tab.cols if c.pk]
    setvals    = addQs(fk.column,' , ')
    whereagain = addQs(fk_pks,' AND ')
    args       = [Arg(get_id_block.name,i) for i in range(len(fk.column))]
    fmt_args2  = [fk_tab.name,setvals,whereagain]
    args.extend(noIndex(fk_pks))

    update_block = SBlock('UPDATE {0} SET {1} WHERE {2}'.format(*fmt_args2)
                         ,'update',args)

    return ins_select + [get_id_block,update_block]

################################################################################
class PBlock(Block):
    """
    A computational block that executes Python code
    """
    def __init__(self
                ,func : Callable
                ,name : Optional[str] = None
                ,args : List[Arg]     = []
                ) -> None:
        assert callable(func), 'Not a func: '+str(func)
        self.func     = Func(func)

        if name is None:
            name = self.func.name

        self.name     = name
        self.args     = args

    def apply(self
             ,curr_dict : Dict[str,Any]
             ,cxn       : Optional[ConnectInfo] = None
             ) -> Any:
        """
        Take a TempFunc's function and wrap it such that it can accept a namespace
            dictionary. The information about how the function interacts with the
            namespace is contained in the TempFunc's args.
        """
        assert cxn is None

        inputvars = [arg.arg_get(curr_dict) for arg in self.args]

        try:
            return self.func.apply(*inputvars)
        except Exception as e:
            msg = '\tApplying func %s in tempfunc %s:\n\t'%(self.func.name,self.name)
            raise ExternalError(msg + format_exc())

# PBlock Shortcuts

def Rename(a:str,b:str)-> Block:
    return PBlock(identity,b,[Arg(a)])

def Unpack(a:str,b:List[str])-> List[Block]:
    return [PBlock(identity,b[i],[Arg(a,i)]) for i in range(len(b))]

def SimpleFunc(func    : Callable
              ,inputs  : List[str] = []
              ,outputs : List[str] = []
              ) -> List[Block]:
    """
    For a rule that has one function in its template
    """
    main_func = PBlock(func,'main',args=[Arg(x) for x in inputs]) # type: Block
    if len(outputs)==1:
        exported = [PBlock(identity,outputs[0],[Arg('main')])] # type: List[Block]
    else:
        exported  = [PBlock(identity,x,[Arg('main',i)])
                        for i,x in enumerate(outputs)]
    return [main_func]+exported

def SimplePipe(funcs   : List[Callable]
              ,inputs  : List[str]
              ,outputs : List[str]
              ,const   : Dict[str,Const] = {}
              ) -> List[Block]:
    """
    A series of functions where the output of each function is fed to the
    following function in the pipeline. You must specify the names of the
    inputs and outputs to the whole process
    """
    init_func = PBlock(funcs[0],args=[Arg(x) for x in inputs])  # type: Block
    remaining = [PBlock(funcs[i],args=[Arg(funcs[i-1].__name__)]) for i in range(1,len(funcs))]   # type: List[Block]
    if len(outputs)==1:
        exported = [PBlock(identity,outputs[0],[Arg(funcs[-1].__name__)])]  # type: List[Block]
    else:
        exported  = [PBlock(identity,x,[Arg(funcs[-1].__name__,i)])
                        for i,x in enumerate(outputs)]
    return  [init_func] + remaining + exported
################################################################################

class Plan(object):
    """

    """
    def __init__(self
                ,blocks : List[Block]
                ,consts : Dict[str,Const] = {}
                ) -> None:
        self.blocks = self.order_blocks(blocks)
        self.consts = consts

    def apply(self,query_dic:Dict[str,Any],cxn:ConnectInfo)->None:
        consts = {k:v.val for k,v in self.consts.items()}
        dic    = {**consts,**query_dic} # MERGE
        for b in self.blocks:
            print('\tapplying '+b.name)
            c = None if isinstance(b,PBlock) else cxn # don't give python cxnInfo
            dic[b.name] = b.apply(dic,c) #type: ignore

    def add_plan(self,sqlite_pth:str,id:int)->None:
        raise NotImplementedError

    @staticmethod
    def order_blocks(bs:List[Block])->List[Block]:
        """
        Prior to storing the list of blocks as a field of the Plan,
        order them such that any block that calls some other block occurs afterwards

        DO WE NEED A HEURISTIC TO HELP WITH INSERT -> SELECT ORDERING?
        OR SHOULD SELECT STATEMENTS ADD "ARTIFICIAL" DEPENDENCIES?
        OR SHOULD S-BLOCKS HAVE AN OPTIONAL 'deps' FIELD WITH SBLOCK NAMES?
        """
        G = DiGraph() # each template has a DAG, with tempfuncs as nodes
        G.add_nodes_from([b.name for b in bs])

        b_dict = {b.name : b for b in bs}

        for b in bs:
            for ba in b.args:
                if ba.name in b_dict.keys():
                    G.add_edge(ba.name,b.name)
            for d in b.deps:
                G.add_edge(d,b.name)
        return topsort_with_dict(G,b_dict)
################################################################################

class Applier(object):
    """
    Class needed for parallel mapping (cannot parallel map lambda functions)
    """
    fail = '%s failed with inputs %s \n%s'

    def __init__(self, rule:'Rule',cxn:ConnectInfo)->None:
        self.rule = rule
        self.cxn  = cxn

    def __call__(self, row:dict)->None:
        try:
            self.rule.plan.apply(row,self.cxn)
        except ExternalError as e:
            row_ = {k:self.abbreviate(v) for k,v in row.items()}
            raise ExternalError(self.fail%(self.rule.name,row_,e))

    @staticmethod
    def abbreviate(x:Any)->Any:
        if isinstance(x,str) and len(x) > 1000:
            return x[:1000]+'...'
        else:
            return x
################################################################################
class Rule(object):
    """
    Information required to transform the database

    Generally, one must provide:
        - a unique name by which this action can be referred (method_name)
        - how to get the required information for the rule (query)
        - a procedure for processing the info and putting it back in the DB (plan)
    """

    def __init__(self
                ,name   : str
                ,query  : Optional[str] = None
                ,plan   : Plan          = Plan([])
                ,stream : bool         = False
                ) -> None:

        if query is not None:
            query = sql_format(query, reindent=True).strip()
            query = sql_format(query, reindent=True) # indempotency

        self.name   = name
        self.query  = query
        self.plan   = plan
        self.stream = stream
        # Precompute dependencies of all SQL commands
        mbQuery = [] if self.query is None else [self.query]
        allsql = mbQuery + [b.text for b in self.plan.blocks if isinstance(b,SBlock)]
        self.deps   = parse_dependencies(allsql)

    def __str__(self)->str:
        return pformat(self.__dict__)

    def __repr__(self)->str:
        return str(self)

    def __eq__(self,other : object)->bool:
        return self.__dict__==other.__dict__

    def dependency_failed(self,conn : ConnectInfo) -> Any:
        q = """SELECT child
                FROM META_dependencies
                JOIN META_rule ON parent = method
                WHERE child =%s AND status like '%%failed%%'"""
        output = sqlselect(conn,q,[self.name])
        return True if len(output) == 0 else output

    def run_rule(self
                ,cxn      : ConnectInfo
                ,parallel : bool = True
                ) -> None:
        """
        Executes a SQL query, then maps each output over a processing function
        which modifies the database.
        """

        try:
            start = time()
            tot   = 0

            if self.query is None:
                inputs = [{}] # type: List[dict]
            else:
                inputs = select_dict(cxn,self.query)  # type: ignore

            if not self.stream:
                inputs = list(inputs)
                l      = len(inputs) # type: Optional[int]
            else:
                l = None
                print("\tDo not know length of input (streaming)")

            if parallel:
                cpus = cpu_count() # -1 or 1 # play safe, leave one free
                with Pool(cpus) as p: # catastrophic failure less likely if we use n-1 available nodes
                    for _ in tqdm(p.imap_unordered(Applier(self,cxn),inputs),length=l):
                        tot+=1
            else:
                for row in tqdm(inputs):
                    Applier(self,cxn)(row)
                    tot+=1

            if not isinstance(inputs,list):
                inputs.close() # type: ignore

            self.update_status(cxn,'completed')
            tot_time = time()-start
            self.update_time(cxn,tot_time/60,safe_div(tot_time,tot))

        except ExternalError as e:
            err = str(e).replace('\\n','\n')
            msg = 'Error when running rule %s\n'%self.name
            print(msg + err)
            self.log_error(cxn,msg + err)

    def log_error(self
                 ,conn : ConnectInfo
                 ,err  : str
                 ) -> None:
        self.update_status(conn,'failed')
        q = "UPDATE META_rule SET error=%s WHERE method=%s"
        sqlexecute(conn,q,[err,self.name])

    def update_status(self
                     ,conn : ConnectInfo
                     ,stat : str
                     ) -> None:
        q = "UPDATE META_rule SET status=%s WHERE method=%s"
        sqlexecute(conn,q,[stat,self.name])

    def update_time(self
                   ,conn     : ConnectInfo
                   ,duration : float
                   ,rate     : float
                   ) -> None:
        q = "UPDATE META_rule SET runtime=%s,rate=%s WHERE method=%s"
        sqlexecute(conn,q,[duration,rate,self.name])

    def add_rule(self,sqlite_pth : str) -> None:
        """
        Insert a Rule to the meta.db
        """
        cmd   = mkInsCmd('rule',['name','query'],sqlite=True)
        sqlite_execute(sqlite_pth,cmd,[self.name,self.query])
        sqlite_execute(sqlite_pth,namespaceCmd,[self.name,'Rule'])
        t_id = handle_to_ind(sqlite_pth,self.name)
        self.plan.add_plan(sqlite_pth,t_id)



# External Modules
from typing import Any
# Internal Modules
from dbgen.support.datatypes.object     import new_model,NOTNULL,FK,UNIQUE,DEFAULT
from dbgen.support.datatypes.sqltypes   import SQLType,Int,Varchar,Text,Decimal
#############################################################################
kris = new_model('kris') # type: Any

class Job(kris):
    """
    DFT jobs
    """
    logfile             = Varchar(),NOTNULL,UNIQUE
    stordir             = Varchar(),NOTNULL
    code                = Varchar(),NOTNULL
    user                = Varchar()
    timestamp           = Int()
    working_directory   = Varchar()
    log                 = Text('long')
    pwinp               = Text()
    potcar              = Text('long')
    poscar              = Text()
    kptcar              = Text()
    paramdict           = Text('long')
    anytraj             = Text('long')
    job_type            = Varchar()
    job_name            = Varchar()
    ads_catalog         = Varchar()
    structure_catalog   = Varchar()

class Calc(kris):
    """
    DFT calc parameters
    """
    dftcode = Varchar(),NOTNULL,UNIQUE
    xc      = Varchar(),NOTNULL,UNIQUE
    pw      = Int(),    NOTNULL,UNIQUE
    psp     = Varchar(),NOTNULL,UNIQUE


class Cell(kris):
    """
    Periodic cells defined by three vectors
    """
    a0              = Decimal(),NOTNULL,UNIQUE
    a1              = Decimal(),NOTNULL,UNIQUE
    a2              = Decimal(),NOTNULL,UNIQUE
    b0              = Decimal(),NOTNULL,UNIQUE
    b1              = Decimal(),NOTNULL,UNIQUE
    b2              = Decimal(),NOTNULL,UNIQUE
    c0              = Decimal(),NOTNULL,UNIQUE
    c1              = Decimal(),NOTNULL,UNIQUE
    c2              = Decimal(),NOTNULL,UNIQUE
    surface_area    = Decimal()
    volume          = Decimal()
    a               = Decimal()
    b               = Decimal()
    c               = Decimal()

class Struct(kris):
    """
    Chemical structure defined in periodic cell
    """
    _components      = [Cell]
    raw              = Text(),   NOTNULL
    rawhash          = Varchar(),NOTNULL
    system_type      = Varchar()
    n_atoms          = Int()
    n_elems          = Int()
    composition      = Varchar()
    composition_norm = Varchar()
    metal_comp       = Varchar()
    str_symbols      = Varchar()
    str_constraints  = Varchar()
    symmetry         = Varchar()
    geo_graph        = Text('long')
    elemental        = Int()

class Element(kris):
    """
    Chemical element
    """
    symbol                  = Varchar(),NOTNULL
    atomic_weight           = Decimal(),NOTNULL
    name                    = Varchar(),NOTNULL
    atomic_radius           = Int()
    phase                   = Varchar()
    group_id                = Int()
    period                  = Int()
    pointgroup              = Varchar()
    spacegroup              = Int()
    evaporation_heat        = Decimal()
    melting_point           = Decimal()
    metallic_radius         = Decimal()
    vdw_radius              = Decimal()
    density                 = Decimal()
    en_allen                = Decimal()
    is_radioactive          = Int()
    lattice_struct          = Varchar()
    fusion_heat             = Decimal()
    econf                   = Varchar()
    covalent_radius_bragg   = Decimal()
    geochemical_class       = Varchar()
    abundance_crust         = Decimal()
    heat_of_formation       = Decimal()
    electron_affinity       = Decimal()
    atomic_volume           = Decimal()
    boiling_point           = Decimal()
    proton_affinity         = Decimal()
    covalent_radius_slater  = Decimal()
    lattice_constant        = Decimal()
    dipole_polarizability   = Decimal()
    en_ghosh                = Decimal()
    thermal_conductivity    = Decimal()
    en_pauling              = Decimal()
    gas_basicity            = Decimal()
    abundance_sea           = Decimal()

class Atom(kris):
    """
    An atom in a specific chemical structure
    """
    _parents     = [Struct]
    _components  = [Element]
    _many_to_one = True

    x            =  Decimal(),NOTNULL
    y            =  Decimal(),NOTNULL
    z            =  Decimal(),NOTNULL
    constrained  = Int(),NOTNULL
    magmom       = Decimal()

class Calc_other(kris):
    """
    Less important DFT Calculator Parameters
    """
    kx          = Int(),NOTNULL
    ky          = Int(),NOTNULL
    kz          = Int(),NOTNULL
    fmax        = Decimal()
    econv       = Decimal(10,7)
    dw          = Int()
    sigma       = Decimal(10,7)
    nbands      = Int()
    mixing      = Decimal()
    nmix        = Int()
    xtol        = Decimal(10,7)
    strain      = Decimal()
    gga         = Varchar()
    luse_vdw    = Int()
    zab_vdw     = Decimal()
    nelmdl      = Int()
    gamma       = Int()
    dipol       = Varchar()
    algo        = Varchar()
    ibrion      = Int()
    prec        = Varchar()
    ionic_steps = Int()
    lreal       = Varchar()
    lvhar       = Int()
    diag        = Varchar()
    spinpol     = Int()
    dipole      = Int()
    maxstep     = Int()
    delta       = Decimal()
    mixingtype  = Varchar()
    bonded_inds = Varchar()
    step_size   = Decimal()
    spring      = Decimal()
    cell_dofree = Varchar()
    cell_factor = Decimal()
    kpts        = Varchar()
    energy_cut_off = Decimal()

class Relax_job(kris):
    """
    Jobs that compute local minima for electronic energy
    """
    _parents    = [Job]
    _components = [Calc,Calc_other]

    reference   = Int()
    is_tom      = Int()

class Roots(kris):
    """
    Directories which are recursively searched for computation logfiles
    """
    root    = Varchar(),NOTNULL,UNIQUE
    code    = Varchar(),NOTNULL,UNIQUE
    label   = Varchar(),NOTNULL,UNIQUE
    active  = Int(),    NOTNULL,UNIQUE,DEFAULT(1)

class Pure_struct(kris):
    """
    Structure abstraction developed by Ankit Jain
    """
    name        = Varchar(),NOTNULL,UNIQUE
    spacegroup  = Int()
    free        = Int()
    nickname    = Varchar()

class Traj(kris):
    """
    A step in a relaxation
    """
    _parents     = [Relax_job]
    _components  = [Struct]
    _many_to_one = True

    final    = Int(),NOTNULL
    energy   = Decimal()
    fmax     = Decimal()
    kptden_x = Decimal()
    kptden_y = Decimal()
    kptden_z = Decimal()

class Adsorbate(kris):
    """
    Species that can adsorb onto a surface
    """
    name        = Varchar(),NOTNULL,UNIQUE
    composition = Varchar()

class Adsorbate_composition(kris):
    """
    Components of an adsorbate
    """
    _parents     = [Adsorbate,Element]

    number = Int(),NOTNULL

class Struct_adsorbate(kris):
    """
    An adsorbate considered on a particular surface
    """
    _parents     = [Adsorbate,Struct]
    _many_to_one = True

    site = Varchar()

#############################################################################
# INSTANCES
#----------
a1 = Atom('1')
a2 = Atom('2')
rj = Relax_job()
vw = kris._mkView([a1.x]) # type: ignore



from typing import Dict,Tuple

def get_kpts_qe(trip:Tuple[str,str,str]) -> Tuple[int,int,int]:
    """
    K point info (? ? ? kx ky kz) - assumed to be in last line of pw.inp
    """

    _,pwinp,_ = trip
    line  = pwinp.strip().split('\n')[-1]
    raw   = line.split()[:3]
    return int(raw[0]),int(raw[1]),int(raw[2])



# External modules
from typing import List
from os         import environ,getenv

# Internal modules
from dbgen.support.datatypes.table     import Table,Col,FK,View
from dbgen.support.datatypes.misc      import Arg,noIndex
from dbgen.support.utils               import addQs,mkInsCmd
from dbgen.core.lists                  import flatten
from dbgen.support.datatypes.sqltypes  import Varchar,Decimal,Text
from dbgen.support.datatypes.rule      import (Rule,SimpleUpsert,SimpleUpdate
                                              ,SimpleInsert,PureSQL,Block,Const
                                              ,SimpleFunc,PBlock,SBlock,Plan
                                              ,SimplePipe)

from dbgen.scripts.IO.get_gpaw_logfiles         import get_gpaw_logfiles
from dbgen.scripts.IO.get_qe_logfiles           import get_qe_logfiles
from dbgen.scripts.IO.get_catalog_logfiles      import get_catalog_logfiles
from dbgen.scripts.IO.get_catalog_sherlock      import get_catalog_sherlock
from dbgen.scripts.IO.get_chargemol_logfiles    import get_chargemol_logfiles
from dbgen.scripts.IO.get_chargemol_sherlock    import get_chargemol_sherlock
from dbgen.scripts.IO.anytraj                   import anytraj
from dbgen.scripts.IO.anytraj_sherlock          import anytraj_sherlock
from dbgen.scripts.IO.parse_mendeleev           import parse_mendeleev
from dbgen.scripts.IO.get_job_metadata          import get_job_metadata
from dbgen.scripts.IO.metadata_sherlock         import metadata_sherlock

from dbgen.scripts.IO.populate_storage          import populate_storage
from dbgen.scripts.IO.storage_sherlock          import storage_sherlock
from dbgen.scripts.Pure.Load.parse_pw_gpaw      import parse_pw_gpaw
from dbgen.scripts.Pure.Load.parse_pw_qe        import parse_pw_qe
from dbgen.scripts.Pure.Load.parse_pw_vasp      import parse_pw_vasp
from dbgen.scripts.Pure.Load.parse_xc_gpaw      import parse_xc_gpaw
from dbgen.scripts.Pure.Load.parse_xc_qe        import parse_xc_qe
from dbgen.scripts.Pure.Load.parse_xc_vasp      import parse_xc_vasp
from dbgen.scripts.Pure.Load.parse_psp_gpaw     import parse_psp_gpaw
from dbgen.scripts.Pure.Load.parse_psp_qe       import parse_psp_qe
from dbgen.scripts.Pure.Load.parse_psp_vasp     import parse_psp_vasp
from dbgen.scripts.Pure.Load.normalize_xc       import normalize_xc
from dbgen.scripts.Pure.Load.normalize_psp      import normalize_psp
from dbgen.scripts.Pure.Atoms.get_traj_gpaw_lazy import get_traj_gpaw_lazy
from dbgen.scripts.Pure.Atoms.get_traj_qe_lazy   import get_traj_qe_lazy
from dbgen.scripts.Pure.Atoms.get_traj_vasp_lazy import get_traj_vasp_lazy
from dbgen.scripts.Pure.Misc.dftcode_dependent  import dftcode_dependent
from dbgen.scripts.Pure.Misc.triple             import triple
from dbgen.scripts.Pure.Misc.identity           import identity
from dbgen.scripts.Pure.Misc.unzip              import unzip
from dbgen.scripts.Pure.Misc.map_               import map_
from dbgen.scripts.Pure.Misc.concat_map         import concat_map
from dbgen.scripts.Pure.Atoms.json_to_traj      import json_to_traj
from dbgen.scripts.Pure.Atoms.traj_to_json      import traj_to_json
from dbgen.scripts.Pure.Atoms.get_cell          import get_cell
from dbgen.scripts.Pure.Atoms.get_atom          import get_atom
from dbgen.scripts.Pure.Atoms.get_system_type   import get_system_type
##########################################################################################
# CONSTANTS
#-----------

calc_funcs = [parse_pw_gpaw,parse_pw_qe,parse_pw_vasp # type: ignore
             ,parse_xc_gpaw,parse_xc_qe,parse_xc_vasp
             ,parse_psp_gpaw,parse_psp_qe,parse_psp_vasp
             ,normalize_xc,normalize_psp,identity]

traj_funcs = [get_cell,get_atom,traj_to_json,get_traj_gpaw_lazy,get_traj_qe_lazy
             ,get_traj_vasp_lazy,identity,unzip]

cell_args = [Arg('cell',i) for i in range(9)]
cell_cols = ['a0','a1','a2','b0','b1','b2','c0','c1','c2']


# Bad stuff
#----------
local =  getenv('SHERLOCK','') != '2'

##########################################################################################
roots = Table('roots'
    ,desc='Directories which are recursively searched for computation logfiles'
    ,cols = [Col('storage_id',      pk=True,auto=True)
            ,Col('root',Varchar(),  nn=True,uniq=True)
            ,Col('code',Varchar(),  nn=True,uniq=True)
            ,Col('label',Varchar(), nn=True,uniq=True)
            ,Col('active',          nn=True,default=1)])

alljob = Table('alljob'
    ,desc='DFT jobs'
    ,cols = [Col('job_id',                          pk=True,auto=True)
            ,Col('logfile',          Varchar(),         nn=True,uniq=True)
            ,Col('stordir',          Varchar(),         nn=True)
            ,Col('code',             Varchar(),         nn=True)
            ,Col('user',             Varchar())
            ,Col('timestamp')
            ,Col('working_directory',Varchar())

            ,Col('log',              Text('long'))
            ,Col('pwinp',            Text())
            ,Col('potcar',           Text('long'))
            ,Col('poscar',           Text())
            ,Col('kptcar',           Text())
            ,Col('paramdict',        Text('long'))
            ,Col('anytraj',          Text('long'))

            ,Col('job_type',         Varchar(),         ind = True)
            ,Col('job_name',         Varchar())
            ,Col('ads_catalog',      Varchar())
            ,Col('structure_catalog',Varchar())
            ,Col('deleted',                         ind=True,nn=True,default=0)])


relax_job = Table('relax_job'
        ,desc = 'Jobs that compute local minima for electronic energy'
        ,cols = [Col('job_id',        pk  = True)
                ,Col('calc_id')
                ,Col('calc_other_id')
                ,Col('reference',     ind = True)
                ,Col('is_tom')]
        ,fks  = [FK('job_id', 'alljob')
                ,FK('calc_id','calc')
                ])#,FK('calc_other_id','calc_other')])

traj = Table('traj'
    ,desc = 'A step in a relaxation'
    ,cols = [Col('job_id',                pk  = True)
            ,Col('traj_id',               pk  = True)
            ,Col('final',                 nn  = True)
            ,Col('struct_id',             nn  = True)
            ,Col('energy',        Decimal(),ind = True)
            ,Col('fmax',          Decimal())
            ,Col('kptden_x',      Decimal())
            ,Col('kptden_y',      Decimal())
            ,Col('kptden_z',      Decimal())]
    ,fks = [FK('job_id',   'relax_job')
           ,FK('struct_id','struct')])

#########################################
calc = Table('calc'
    ,desc='DFT calc parameters'
    ,cols = [Col('calc_id',          pk = True,auto = True)
            ,Col('dftcode',   Varchar(), nn = True,uniq = True)
            ,Col('xc',        Varchar(), nn = True,uniq = True)
            ,Col('pw',                   nn = True,uniq = True)
            ,Col('psp',       Varchar(), nn = True,uniq = True)])

element = Table('element'
    ,desc = 'ID = atomic number' # table for storing data in element.json
    ,cols = [Col('element_id',                     pk = True)
            ,Col('symbol',                 Varchar(),  nn = True)
            ,Col('atomic_weight',          Decimal(),nn = True)
            ,Col('name',                   Varchar(),  nn = True)
            ,Col('atomic_radius')
            ,Col('phase',                  Varchar())
            ,Col('group_id')
            ,Col('period')
            ,Col('pointgroup',             Varchar())
            ,Col('spacegroup')
            ,Col('evaporation_heat',        Decimal())
            ,Col('melting_point',           Decimal())
            ,Col('metallic_radius',         Decimal())
            ,Col('vdw_radius',              Decimal())
            ,Col('density',                 Decimal())
            ,Col('en_allen',                Decimal())
            ,Col('is_radioactive')
            ,Col('lattice_struct',          Varchar())
            ,Col('fusion_heat',             Decimal())
            ,Col('econf',                   Varchar())
            ,Col('covalent_radius_bragg',   Decimal())
            ,Col('geochemical_class',       Varchar())
            ,Col('abundance_crust',         Decimal())
            ,Col('heat_of_formation',       Decimal())
            ,Col('electron_affinity',       Decimal())
            ,Col('atomic_volume',           Decimal())
            ,Col('boiling_point',           Decimal())
            ,Col('proton_affinity',         Decimal())
            ,Col('covalent_radius_slater',  Decimal())
            ,Col('lattice_constant',        Decimal())
            ,Col('dipole_polarizability',   Decimal())
            ,Col('en_ghosh',                Decimal())
            ,Col('thermal_conductivity',    Decimal())
            ,Col('en_pauling',              Decimal())
            ,Col('gas_basicity',            Decimal())
            ,Col('abundance_sea',           Decimal())])
#########
cell = Table('cell'
    ,desc = 'Periodic cells defined by three vectors'
    ,cols = [Col('cell_id',             pk = True,auto = True)
            ,Col('a0',          Decimal(),nn = True,uniq = True)
            ,Col('a1',          Decimal(),nn = True,uniq = True)
            ,Col('a2',          Decimal(),nn = True,uniq = True)
            ,Col('b0',          Decimal(),nn = True,uniq = True)
            ,Col('b1',          Decimal(),nn = True,uniq = True)
            ,Col('b2',          Decimal(),nn = True,uniq = True)
            ,Col('c0',          Decimal(),nn = True,uniq = True)
            ,Col('c1',          Decimal(),nn = True,uniq = True)
            ,Col('c2',          Decimal(),nn = True,uniq = True)
            ,Col('surface_area',Decimal())
            ,Col('volume',      Decimal())
            ,Col('a',           Decimal())
            ,Col('b',           Decimal())
            ,Col('c',           Decimal())])

struct = Table('struct'
    ,desc = 'List of unique combinations of [atom] and cell'
    ,cols = [Col('struct_id',                pk  = True,auto = True)
            ,Col('cell_id',                  nn  = True)
            ,Col('raw',             Text(),  nn  = True)
            ,Col('rawhash',         Varchar(),   nn  = True,uniq = True)
            ,Col('system_type',     Varchar(),   ind = True) # used to partition subclasses
            ,Col('n_atoms')
            ,Col('n_elems')
            ,Col('composition',     Varchar())
            ,Col('composition_norm',Varchar())
            ,Col('metal_comp',      Varchar())
            ,Col('str_symbols',     Varchar(510))
            ,Col('str_constraints', Varchar())
            ,Col('symmetry',        Varchar())
            ,Col('pure_struct_id')
            ,Col('pure_struct_id_catalog')
            ,Col('species_id')
            ,Col('spacegroup')
            ,Col('geo_graph',       Text('long'))
            ,Col('elemental')]
    ,fks  = [FK('cell_id',       'cell')
            ])#,FK('pure_struct_id','pure_struct')
            #,FK('species_id','species')])

atom = Table('atom'
    ,desc = 'List of atoms in unique structs'
    ,cols = [Col('struct_id',                pk = True)
            ,Col('atom_id',                  pk = True)
            ,Col('x',            Decimal(),  nn = True)
            ,Col('y',            Decimal(),  nn = True)
            ,Col('z',            Decimal(),  nn = True)
            ,Col('constrained')
            ,Col('magmom',       Decimal())
            ,Col('tag')
            ,Col('v3')
            ,Col('v4')
            ,Col('v5')
            ,Col('v6')
            ,Col('v7')
            ,Col('v8')
            ,Col('v9')
            ,Col('v10')
            ,Col('adsorbate_id')
            ,Col('struct_adsorbate_id')
            ,Col('adsorbate_id_catalog')
            ,Col('struct_adsorbate_id_catalog')
            ,Col('layer')]
    ,fks = [FK('struct_id', 'struct')
           ,FK('element_id','element')
           ])#,FK(['struct_id'
             #  ,'adsorbate_id'
             # ,'struct_adsorbate_id']
             # ,'struct_adsorbate')])


trajatom = Table('trajatom'
        ,desc = "An atom considered as part of an optimization trajectory"
        ,cols = [Col('job_id',        pk=True)
                ,Col('traj_id',        pk=True)
                ,Col('struct_id',    pk=True)
                ,Col('atom_id',        pk=True)
                ,Col('fx',Decimal())
                ,Col('fy',Decimal())
                ,Col('fz',Decimal())]
        ,fks  = [FK(['job_id','traj_id'],'traj')
                ,FK(['struct_id','atom_id'],'atom')])
tables = [roots,alljob,relax_job,traj,calc,element,cell,struct,atom,trajatom]
##########################################################################################
##########################################################################################
##########################################################################################
job = View('job',"""SELECT A.job_id,A.user,A.timestamp,A.logfile,A.stordir,A.code
                          ,A.job_type,A.working_directory,A.ads_catalog,A.structure_catalog
                          ,A.deleted,A.job_name
                    FROM alljob AS A
                    WHERE NOT A.deleted""")
finaltraj = View('finaltraj','SELECT * FROM traj WHERE traj.final')


views = [job,finaltraj]
##########################################################################################
##########################################################################################
##########################################################################################

populate_gpaw_job = Rule(name = 'stray_gpaw'
    ,query    = """SELECT R.root
                    FROM roots AS R
                    WHERE R.code = 'gpaw' AND R.label = 'stray' AND R.active
                    GROUP BY R.root"""
    ,plan = Plan(blocks = SimpleFunc(get_gpaw_logfiles
                                    ,['root'],['stordir','logfile'])
                         +SimpleInsert(alljob,['stordir','logfile','code'])
                ,consts={'code':Const('gpaw')}))

populate_qe_job = Rule(name = 'stray_qe'
    ,query    = """SELECT R.root
                    FROM roots AS R
                    WHERE R.code = 'quantumespresso'
                      AND R.label = 'stray'
                      AND R.active"""
    ,plan = Plan(blocks = SimpleFunc(get_qe_logfiles
                                    ,['root'] ,['stordir','logfile'])
                         + SimpleInsert(alljob,['stordir','logfile','code'])
                ,consts={'code':Const('quantumespresso')}))

pop_job_cols = ['stordir','logfile','code','log','pwinp','potcar','poscar','kptcar'
               ,'paramdict']

populate_catalog_job = Rule(name = 'catalog'
    ,query    = """SELECT  R.root
                          ,(SELECT GROUP_CONCAT(stordir) FROM alljob) AS existing
                    FROM roots AS R
                    WHERE R.label = 'catalog'
                      AND R.active
                    """
    ,plan = Plan(SimpleFunc(get_catalog_logfiles if local else get_catalog_sherlock # type: ignore
                          ,['root'] if local else ['root','existing'],pop_job_cols)
                +SimpleInsert(alljob,pop_job_cols)))

populate_chgmol_job = Rule(name = 'stray_chargemol'
    ,query    = """SELECT  R.root
                    FROM roots AS R
                    WHERE R.code = 'chargemol'
                        AND R.label = 'stray'
                        AND R.active
                    """
    ,plan = Plan(blocks = SimpleFunc(get_chargemol_logfiles if local else get_chargemol_sherlock
                                      ,['root'] ,['stordir','logfile'])
                         +SimpleInsert(alljob,['stordir','logfile','code'])
                ,consts = {'code':Const('chargemol')}))

anyTraj = Rule(name='anytraj'
    ,query    = """SELECT J.stordir,J.job_id FROM alljob AS J
                    WHERE J.anytraj is NULL AND NOT J.deleted"""
    ,plan = Plan(SimpleFunc(anytraj if local else anytraj_sherlock,['stordir'],['anytraj'])
                +[SimpleUpdate(alljob,['anytraj'])]))

# ################################################################################

mdata_cols = ['user','timestamp','working_directory'
             ,'job_type','job_name']

getJobMetadata = Rule(name = 'metadata'
    ,query    = """SELECT J.job_id,J.stordir,J.code,J.paramdict
                    FROM alljob AS J
                    WHERE J.user is null AND NOT J.deleted"""
    ,plan = Plan(SimpleFunc(get_job_metadata if local else metadata_sherlock
                          ,['stordir','code','paramdict'],mdata_cols)
                + [SimpleUpdate(alljob,mdata_cols)]))

# ################################################################################
calc_blocks = [PBlock(dftcode_dependent
              ,name = x
              ,args = noIndex(['dftcode'
                              ,'log'
                              ,'parse_%s_gpaw'%x
                              ,'parse_%s_qe'%x
                              ,'parse_%s_vasp'%x
                              ,'identity'
                              ,'identity' if x =='pw' else 'normalize_%s'%x]))
              for x in ['pw','xc','psp']] # type: List[Block]
populate_relax_calc = Rule(name='calc'
        ,query    = """SELECT relax_job.job_id,J.log,J.code AS dftcode
                        FROM relax_job
                        JOIN alljob as J USING (job_id)
                        WHERE relax_job.calc_id is null
                             AND NOT J.deleted"""
        ,plan = Plan(blocks = calc_blocks
                            + SimpleUpsert(ins_tab   = calc
                                        ,ins_cols  = ["dftcode","xc","pw","psp"]
                                        ,fk_tab    = relax_job)
                    ,consts = {f.__name__:Const(f) for f in calc_funcs}))

populate_relax_job = Rule(name = 'relax_job'
    ,plan = Plan(PureSQL(["""
        INSERT INTO relax_job (job_id)
        SELECT J.job_id FROM alljob AS J
        WHERE J.job_type IN ('relax','latticeopt','vcrelax')
            AND J.job_id NOT IN (SELECT R.job_id from relax_job AS R)
            AND NOT J.deleted"""])))

# if any ionic steps are in the traj table, then we assume all 4 tables have been taken care of
populate_traj = Rule(name='traj'
    ,query = """SELECT R.job_id,J.code AS dftcode,J.log,J.anytraj,J.poscar,J.stordir
                 FROM relax_job AS R JOIN alljob AS J USING (job_id)
                 WHERE R.job_id NOT IN (SELECT traj.job_id FROM traj)
                 AND J.code != 'vasp' # TEMPORARY THING
                 AND NOT J.deleted"""

    ,plan = Plan(consts = {f.__name__:Const(f) for f in traj_funcs}
        ,blocks=[
              PBlock(triple
                      ,'anytraj_log_poscar'
                      ,args = noIndex(['anytraj','log','poscar']))


              ,PBlock(dftcode_dependent
                       ,name = 'get_traj'
                       ,args = noIndex(['dftcode'
                                  ,'anytraj_log_poscar'
                                  ,'get_traj_gpaw_lazy'
                                  ,'get_traj_qe_lazy'
                                  ,'get_traj_vasp_lazy'
                                  ,'identity'
                                  ,'unzip'])) # returns ([ASEATOMS]m,[STEPNUM]m,[ENERGY]m,[TRAJATOMLIST]m)
               ,PBlock(map_
                        ,name = 'cell'
                        ,args = noIndex(['get_cell','structs_few'])) # returns ([ax]m,[ay]m,...)
               ,PBlock(map_
                        ,name = 'structs_few'
                        ,args = [Arg('traj_to_json')
                                ,Arg('get_traj', 0)]) # returns [raw_json]m
               ,PBlock(concat_map
                        ,name = 'atoms'
                        ,args = noIndex(['get_atom','structs_few'])) # returns ([ASEATOMS]n,[Ind]n,[ELEMENT]n,[X]n,[Y]n,...)
               ,PBlock(identity
                        ,name = 'structs_many'
                        ,args = [Arg('atoms', 0)])# returns [raw_json]n
                ,PBlock(flatten
                         ,name = 'flattened_trajatoms'
                         ,args = [Arg('get_traj',3)])  # [STEPNUM,Ind,FX,FY,FZ]n
                ,PBlock(unzip
                         ,'trajatoms'
                         ,args=[Arg('flattened_trajatoms')]) # ([STEPNUM]n,[Ind]n,[FX]n,[FY]n,[FZ]n)

        ,SBlock(name = 'insert_cell'
             ,text = mkInsCmd('cell',cell_cols,sqlite=False)
             ,args = cell_args)

       ,SBlock(name = 'query_cell'
             ,text = ('SELECT C.cell_id FROM cell AS C WHERE '
                     +addQs(['C.%s'%x for x in cell_cols],' AND '))
             ,args = cell_args, deps=['insert_cell'])

       ,SBlock(name = 'insert_struct'
             ,text = """INSERT INTO struct (raw,rawhash,cell_id)
                           VALUES (%s,SHA2(%s,512),%s)
                        ON DUPLICATE KEY UPDATE cell_id=cell_id"""
             ,args = [Arg('structs_few')
                      ,Arg('structs_few')
                     ,Arg('query_cell',0)])

       ,SBlock(name = 'query_struct_few'
             ,text = 'SELECT S.struct_id FROM struct AS S WHERE S.rawhash = SHA2(%s,512)'
             ,args = [Arg('structs_few')],deps=['insert_struct'])

       ,SBlock(name = 'query_struct_many'
             ,text = 'SELECT S.struct_id FROM struct AS S WHERE S.rawhash = SHA2(%s,512)'
             ,args = [Arg('structs_many')],deps=['insert_struct'])

       ,SBlock(name = 'insert_atom'
             ,text = mkInsCmd('atom',['struct_id','atom_id','element_id'
                                  ,'x','y','z','constrained','magmom','tag']
                            ,sqlite=False)
             ,args = [Arg('query_struct_many')]
                    +[Arg('atoms',i) for i in range(1,9)])

     ,SBlock(name = 'insert_traj'
           ,text = mkInsCmd('traj',['job_id','traj_id','struct_id','energy','final'],sqlite=False)
           ,args = [Arg('job_id')
                   ,Arg('get_traj',1)
                   ,Arg('query_struct_few')
                   ,Arg('get_traj',2)
                   ,Arg('get_traj',4)])

    ,SBlock(name = 'insert_trajatom'
          ,text = mkInsCmd('trajatom',['job_id','traj_id','struct_id','atom_id','fx','fy','fz'],sqlite=False)
          ,args = [Arg('job_id')
                  ,Arg('trajatoms',0)
                  ,Arg('query_struct_many')]
                 +[Arg('trajatoms',i) for i in range(1,5)])]))

allcols = ['element_id', 'symbol', 'name', 'atomic_weight','atomic_radius'
          , 'phase','evaporation_heat', 'pointgroup','spacegroup'
          , 'melting_point', 'metallic_radius', 'vdw_radius'
          , 'density', 'en_allen', 'is_radioactive'
          , 'lattice_struct' , 'fusion_heat'
          , 'econf', 'period', 'covalent_radius_bragg'
          , 'geochemical_class', 'abundance_crust', 'heat_of_formation'
          , 'electron_affinity', 'atomic_volume',  'boiling_point'
          , 'proton_affinity', 'covalent_radius_slater'
          , 'lattice_constant', 'dipole_polarizability'
          , 'en_ghosh', 'thermal_conductivity', 'group_id', 'en_pauling'
          , 'gas_basicity','abundance_sea']

populate_element = Rule(name='element'
    ,plan = Plan(SimpleFunc(parse_mendeleev,[],allcols)
            +SimpleInsert(element,allcols)))

sys_type = Rule(name='systype'
    ,query      = """SELECT S.struct_id,S.raw
                     FROM struct AS S
                     WHERE S.system_type IS NULL"""
    ,plan   = Plan(SimplePipe([json_to_traj,get_system_type]
                       ,inputs  = ['raw']
                       ,outputs = ['system_type'])
     + [SimpleUpdate(struct,['system_type'])]))


cell_info = Rule(name='cell'
    ,plan=Plan(PureSQL(["""UPDATE cell AS C SET surface_area = SQRT((C.a0*C.b2-C.a2*C.b1) * (C.a1*C.b2 - C.a2*C.b1)
                                                                 +(C.a0*C.b2-C.a2*C.b0) * (C.a0*C.b2 - C.a2*C.b0)
                                                                 +(C.a0*C.b1-C.a1*C.b0) * (C.a0*C.b1 - C.a1*C.b0))
                                              ,volume = C.a0*(C.b1*C.c2-C.b2*C.c1)
                                                       -C.a1*(C.b0*C.c2-C.b2*C.c0)
                                                       +C.a2*(C.b0*C.c1-C.b1*C.c0)
                                              ,a = SQRT(C.a0*C.a0+C.a1*C.a1+C.a2*C.a2)
                                              ,b = SQRT(C.b0*C.b0+C.b1*C.b1+C.b2*C.b2)
                                              ,c = SQRT(C.c0*C.c0+C.c1*C.c1+C.c2*C.c2)
                        WHERE C.cell_id > 0"""])))
n_atoms = Rule(name='n_atoms'
    ,plan=Plan(PureSQL(["""UPDATE struct SET n_atoms = (SELECT count(1)
                                FROM atom WHERE atom.struct_id=struct.struct_id
                                GROUP BY atom.struct_id)
                          WHERE struct_id > 0"""

                    ,"""UPDATE struct SET n_elems = (SELECT count(1)
                                    FROM struct_composition AS C WHERE C.struct_id=struct.struct_id
                                    GROUP BY C.struct_id)
                                    WHERE struct_id > 0"""])))

stor_cols = ['root','code','label']
pop_storage= Rule('root'
    ,plan = Plan(SimpleFunc(populate_storage if local else storage_sherlock
                           ,outputs=stor_cols)
                +SimpleInsert(roots,stor_cols)))

rules = [pop_storage
        ,populate_gpaw_job
        ,populate_qe_job
         ,populate_catalog_job
        ,populate_chgmol_job
        ,getJobMetadata
        ,populate_relax_calc
        ,populate_relax_job
        ,populate_traj
        ,populate_element
        ,sys_type
        ,cell_info
        ,n_atoms
        ,anyTraj]



#External Modules
from typing import List,Dict,Type,TYPE_CHECKING
if TYPE_CHECKING:
    from dbgen.support.datatypes.object import Object

from abc    import abstractmethod
#Internal Modules
from dbgen.support.datatypes.attr       import Attr
from dbgen.support.datatypes.rule       import Block,SBlock,PBlock,SimpleUpsert
from dbgen.support.datatypes.misc       import noIndex,Arg
from dbgen.support.utils                import addQs
################################################################################

class InsertUpdate(object):
    attrs = [] # type: List[Attr]
    @abstractmethod
    def mk_block(self)->List[Block]:
        raise NotImplementedError

class Insert(InsertUpdate):
    """
    Take items in a namespace and store them as attributes on objects
    in the database
    """
    def __init__(self
                ,uni    : Dict[str,'Object']
                ,attrs  : List[Attr]
                ) -> None:
        self.obj = uni[attrs[0].obj]
        assert all([a.obj == self.obj._name for a in attrs])
        self.attrs = attrs
        self.n     = len(attrs)
        self.colnames = [a.name for a in self.attrs]
        self.args  = noIndex(self.colnames)

    def mk_block(self)->List[Block]:
        """

        """
        qs       = ','.join(['%s']*self.n)
        name     = 'insert_%s_%s'%(self.obj._name,'-'.join(self.colnames))
        text     = 'INSERT INTO %s (%s) VALUES (%s)'%(self.obj._name
                                                   ,','.join(self.colnames)
                                                   ,qs)
        dup      = ' ON DUPLICATE KEY UPDATE {0} = {0}'.format(self.colnames[0])

        return [SBlock(name = name
                     ,text = text + dup
                     ,args = self.args)
                ,self.mkSelectBlock()]

    def mkSelectBlock(self)->Block:
        """
        Given any insertion, we can construct a block that queries which ID
        """
        where    = addQs([str(a) for a in self.attrs],' AND ')
        fmt_args = [','.join(map(str,self.obj._pks)),self.obj._name,where]

        return SBlock(name = 'select_'+self.obj._name  # type: ignore
                     ,text = 'SELECT {0} FROM {1} WHERE {2}'.format(*fmt_args)
                     ,args = self.args)


class Update(InsertUpdate):
    """
    Update attributes of an object
    """
    def __init__(self
                ,uni   : Dict[str,'Object']
                ,attrs : List[Attr]
                ) -> None:
        self.obj = uni[attrs[0].obj]                        # type: ignore
        assert all([a.obj == self.obj._name for a in attrs]) # type: ignore
        self.attrs = attrs

    def mk_block(self)->List[Block]:
        colnames = [a.name for a in self.attrs]
        pk_cols  = [a.name for a in self.obj._pks]  # type: ignore
        name     = 'update_%s_%s'%(self.obj._name,','.join(colnames))   # type: ignore
        setstr   = addQs(colnames,',')
        where    = addQs(pk_cols,' AND ')
        text     = 'UPDATE %s SET %s WHERE %s'%(self.obj._name,setstr,where)   # type: ignore
        args = noIndex(colnames+pk_cols)
        return [SBlock(name = name
                    ,text = text
                    ,args = args)]

class InsertFK(InsertUpdate):
    def __init__(self
                ,uni      : Dict[str,'Object']
                ,ins_cols : List[Attr]
                ,fk_tab   : Type['Object']
                ) -> None:
        self.uni      = uni
        self.ins_cols = ins_cols
        self.fk_tab   = fk_tab

    def mk_block(self)->List[Block]:

        ins_obj = self.uni[self.ins_cols[0].obj]
        cols    = [x.name for x in self.ins_cols]
        return SimpleUpsert(ins_obj._mkTable(),cols,self.fk_tab._mkTable()) # type: ignore
        # fk      = [f for f in self.fk_tab._fks if f._to is ins_obj][0]   # type: ignore
        # fk_cols = fk.from_attr
        # pk_cols = fk.to_attr
        # n       = len(fk_cols)
        # insert_block = Insert(self.uni,self.ins_cols).mk_block()[0]
        #
        # where    = addQs(map(str,self.ins_cols),' AND ')                # type: ignore
        # fmt_args = [','.join(map(str,pk_cols)),ins_obj._name,where]     # type: ignore
        #
        #
        # get_id_block = SBlock(text = 'SELECT {0} FROM {1} WHERE {2}'.format(*fmt_args)
        #                     ,name = 'select_'+self.name
        #                     ,args = insert_block.args)
        #
        # update_block = Update(self.uni,fk_cols).mk_block()[0]
        # update_block.args[:n] = [Arg('select_'+self.name,i) for i in range(n)]
        #
        # return [insert_block, get_id_block, update_block]



from typing     import Tuple,List
from os         import environ
from subprocess import check_output

def storage_sherlock()->Tuple[List[str],List[str],List[str]]:
    """
    Reads DFT_LOGFILES to find a structured textfile with information
    about what directories should be considered 'roots'
    """
    path = '/scratch/users/ksb/share/reference/roots.txt'
    with open(path,'r') as f:
        rootdata  = [l.split()
                        for l in f.readlines()
                        if '#' not in l and l.strip()!='']
    paths,codes,labels = zip(*rootdata)
    return list(paths),list(codes),list(labels)



# External modules
from ase.io     import read     # type: ignore
from glob       import glob
# Internal modules
from dbgen.scripts.Pure.Atoms.traj_to_json import traj_to_json
################################################################################
def anytraj_sherlock(root : str) -> str:
    """
    ASE IO read function - takes any traj it can find
    """
    trajs = glob('%s/*.traj'%root)
    if len(trajs) == 0:
        raise ValueError('Get Traj could not find any traj in '+root)
    else:
        atoms = read(trajs[0])
        return traj_to_json(atoms)

if __name__=='__main__':
    import sys
    print(anytraj_sherlock(sys.argv[1]))



from typing import List,FrozenSet,Dict
from dbgen.support.datatypes.relation   import Relation
from dbgen.support.datatypes.func       import Func
from dbgen.support.datatypes.dbg        import DBG
from dbgen.support.datatypes.misc       import ConnectInfo

################################################################################
class Model(object):
    def __init__(self
                ,schema    : Dict[str,type]
                ,relations : List[Relation]
                ) -> None:
        self.schema    = schema
        self.relations = relations
        self.objects   = schema.values()

    def createDBG(self)->DBG:
        tables = [o._mkTable() for o in self.objects]  # type: ignore
        rules  = [r.mk_rule() for r in self.relations]
        return DBG(tables=tables,rules=rules)

    def run(self
           ,reset   : bool      = False
           ,catalog : bool      = False
           ,add     : bool      = False
           ,xclude  : str       = ''
           ,only    : str       = ''
           ) -> None:
        db = ConnectInfo()

        onl = set(only.split())

        x  = set() if catalog else {'catalog'}
        self.createDBG().run_all(db
                                ,reset    = reset
                                ,xclude   = x | set(xclude.split())
                                ,add      = add
                                ,parallel = False
                                ,only     = onl)




#Typing imports
import typing as typ

#External imports
from os             import environ
from argparse       import ArgumentParser
from pprint         import pformat
from distutils.util import strtobool
from ast            import literal_eval
from json           import load

# Internal Modules
from dbgen.support.utils   import ConnectInfo
from dbgen.core.misc       import levenshteinDistance
from dbgen.inputs.main     import dbg
"""
Command line interface for executing a database update

Requires DB_JSON environment variable to specify database that will be modified

The following flags control program behavior:
    - verbose
    - local
    - reset
    - add
    - meta
    - only
    - xclude
    - jobs
    - catalog
    - test
    - parallel

"""
#Internal Imports
import parser # type: ignore

class CLICommand:
    short_description = "Command line interface for executing a database update"
    description = ("""
The following flags control program behavior:
        - verbose
        - local
        - reset
        - add
        - meta
        - only
        - xclude
        - jobs
        - catalog
        - test
        - parallel""")

    @staticmethod
    def add_arguments(parser : 'parser') -> None: # type: ignore
        parser.add_argument('--verbose'
                           ,default = False
                           ,type    = lambda x:bool(strtobool(x))
                           ,help    = """Leave it off please - hard to find a reasonable
                                        amount of things that's neither too little nor
                                        too much ... maybe the input for this should
                                        actually be 'high 'low' or 'medium'?""")

        parser.add_argument('--local'
                           ,default = False
                           ,type    = lambda x:bool(strtobool(x))
                           ,help    = 'Default: if on sherlock, don\'t run local')

        parser.add_argument('--reset'
                           ,default = False
                           ,type    = lambda x:bool(strtobool(x))
                           ,help    = 'Nuke the DB first - needed if you make schema changes')

        parser.add_argument('--add'
                           ,default = False
                           ,type    = lambda x:bool(strtobool(x))
                           ,help    = 'Try to add columns (if you make benign schema additions)')

        parser.add_argument('--meta'
                           ,default = False
                           ,type    = lambda x:bool(strtobool(x))
                           ,help    = 'Make a meta.db')

        parser.add_argument('--only'
                           ,default = ''
                           ,help    = 'Run only the (space separated) rules')

        parser.add_argument('--xclude'
                           ,default = 'underlying ads_triple'
                           ,help    = 'Do not run the (space separated) rules')

        parser.add_argument('--jobs'
                           ,default = False
                           ,type    = lambda x:bool(strtobool(x))
                           ,help    = 'Run all job_loading rules (takes time to rerun even if no new jobs)')

        parser.add_argument('--catalog'
                           ,default = False
                           ,type    = lambda x:bool(strtobool(x))
                           ,help    = 'Run only the catalog job loading rule ')

        parser.add_argument('--parallel'
                           ,default = True
                           ,type    = lambda x:bool(strtobool(x))
                           ,help    = 'Parallelize processing of multiple Query inputs. (Turn off if using pdb)')

    @staticmethod
    def run(args : typ.Any) -> None:
        """
        Executes DB modifying procedure described by a DBG instance
        """
        print('args are: \n'+pformat(args))


        # Actually interact with database
        #------------------------------------
        if args.local:
            db = ConnectInfo()
        else:
            with open(environ['DB_JSON'],'r') as f:
                db = ConnectInfo(**load(f))

        if args.meta:
            dbg.test_meta() # create a meta-db

        # Determine which job-loading rules to include
        #------------------------------------------------------------
        if not args.jobs:
            args.xclude+=''' stray_gpaw
                             stray_qe
                             stray_chargemol ''' # these T's ONLY run if --jobs

        if (not args.reset) and (not (args.catalog or args.jobs)):
            args.xclude+=''' catalog ''' # this run if EITHER --jobs OR --catalog...provided we didn't just nuke DB

        # Process string inputs
        #------------------------
        only   = set(args.only.split())
        xclude = set(args.xclude.split()) - only

        # Verify every  rule excluded/included by name actually exists
        # if first 5 chars match or within edit distance of 5, suggest alternatives
        #------------------------------------------------------------------------------------
        for w in (only | xclude):
            match = False
            close = []
            for n in [t.name for t in dbg.rules]:
                d = levenshteinDistance(w,n)
                upW,upN = max(len(w),5), max(len(n),5) # variables for safe indexing
                if d == 0:
                    match = True
                    break
                elif d < 5 or w[:upW] == n[:upN]:
                    close.append(n) # keep track of near-misses
            if not match:
                did_you = "Did you mean %s"%close if close else ''
                raise ValueError("No match found for %s\n%s"%(w,did_you))

        dbg.run_all(db
                   ,only    = only
                   ,xclude  = xclude
                   ,reset   = args.reset
                   ,add     = args.add
                   ,verbose = args.verbose
                   ,parallel= args.parallel)



# External Modules
from os             import environ
from argparse       import ArgumentParser
from pprint         import pformat
from distutils.util import strtobool
from ast            import literal_eval
from json           import load

# Internal Modules
from dbgen.support.utils   import ConnectInfo
from dbgen.core.misc       import levenshteinDistance
from dbgen.inputs.main     import dbg
"""
Command line interface for executing a database update

Requires DB_JSON environment variable to specify database that will be modified

The following flags control program behavior:
    - db
    - local
    - reset
    - add
    - meta
    - only
    - xclude
    - jobs
    - catalog
    - test
    - parallel
"""

################################################################################
##########
# Settings
#---------
parser = ArgumentParser(description  = 'Run a DBG update'
                       ,allow_abbrev = True)

parser.add_argument('--db'
                   ,default = ''
                   ,help    = "Override the schema defined in DB_JSON")

parser.add_argument('--local'
                   ,default = False
                   ,type    = lambda x:bool(strtobool(x))
                   ,help    = 'Default: if on sherlock, don\'t run local')

parser.add_argument('--reset'
                   ,default = False
                   ,type    = lambda x:bool(strtobool(x))
                   ,help    = 'Nuke the DB first - needed if you make schema changes')

parser.add_argument('--add'
                   ,default = False
                   ,type    = lambda x:bool(strtobool(x))
                   ,help    = 'Try to add columns (if you make benign schema additions)')

parser.add_argument('--meta'
                   ,default = False
                   ,type    = lambda x:bool(strtobool(x))
                   ,help    = 'Make a meta.db')

parser.add_argument('--only'
                   ,default = ''
                   ,help    = 'Run only the (space separated) rules')

parser.add_argument('--xclude'
                   ,default = 'underlying ads_triple'
                   ,help    = 'Do not run the (space separated) rules')

parser.add_argument('--jobs'
                   ,default = False
                   ,type    = lambda x:bool(strtobool(x))
                   ,help    = '''Run all job_loading rules
                                (takes time to rerun even if no new jobs)''')

parser.add_argument('--catalog'
                   ,default = False
                   ,type    = lambda x:bool(strtobool(x))
                   ,help    = 'Run the catalog job loading rule ')

parser.add_argument('--parallel'
                   ,default = True
                   ,type    = lambda x:bool(strtobool(x))
                   ,help    = '''Parallelize processing of multiple Query inputs.
                                 (Turn off if using pdb)''')
################################################################################
def load_jobs(jobs     : bool = False
             ,catalog  : bool = True
             ,add      : bool = False
             ,reset    : bool = False
             ,parallel : bool = True
             ,only     : str  = ''
             ,xclude   : str  = ''
             ) -> None:
        """
        Calls the main function from within Python (bypass command line args)
        See parser description for input variable meanings
        """

        main({'jobs':jobs,'catalog':catalog,'add':add,'reset':reset
             ,'parallel':parallel,'only':only,'xclude':xclude})

def main(direct_args : dict = {})->None:
    """
    Executes DB modifying procedure described by a DBG instance
    """
    # Get args
    #------------
    parser.set_defaults(**direct_args)
    args = parser.parse_args() # get args from command line
    print('args are: \n'+pformat(args))

    if args.local:
        db = ConnectInfo()
    else:
        with open(environ['DB_JSON'],'r') as f:
            db = ConnectInfo(**load(f))

    if args.db:
        db.db = args.db

    if args.meta:
        dbg.test_meta() # create a meta-db

    # Determine which job-loading rules to include
    #------------------------------------------------------------
    if not args.jobs:
        args.xclude+=''' stray_gpaw
                         stray_qe
                         stray_chargemol ''' # these T's ONLY run if --jobs

    if (not args.reset) and (not (args.catalog or args.jobs)):
        args.xclude+=''' catalog ''' # this run if EITHER --jobs OR --catalog...provided we didn't just nuke DB

    # # Process string inputs
    #------------------------
    only   = set(args.only.split())
    xclude = set(args.xclude.split()) - only

    # Verify every rule excluded/included by name actually exists
    # if first 5 chars match or within edit distance of 5, suggest alternatives
    #------------------------------------------------------------------------------------
    for w in (only | xclude):
        match = False
        close = []
        for n in [t.name for t in dbg.rules]:
            d = levenshteinDistance(w,n)
            upW,upN = max(len(w),5), max(len(n),5) # variables for safe indexing
            if d == 0:
                match = True
                break
            elif d < 5 or w[:upW] == n[:upN]:
                close.append(n) # keep track of near-misses
        if not match:
            did_you = "Did you mean %s"%close if close else ''
            raise ValueError("No match found for %s\n%s"%(w,did_you))

    dbg.run_all(db
               ,only    = only
               ,xclude  = xclude
               ,reset   = args.reset
               ,add     = args.add
               ,parallel= args.parallel)

if __name__ == '__main__':
    main()



from typing     import List,Tuple
from os         import listdir
from glob       import glob
################################################################################
def get_catalog_sherlock(rootpath : str
                        ,existing : str
                    ) -> Tuple[List[str],List[str],List[str]
                              ,List[str],List[str],List[str]
                              ,List[str],List[str],List[str]]:
    """
    Searches for DFT calculation folders. Returns unzipped (stordir,logfile,code) triples

    to get anytraj: \$(cat \$(ls \$f/*.traj | head -1))
    """
    ######################
    # Initialize Variables
    #---------------------
    filt  = set(existing.split(","))
    files = set(listdir(rootpath))
    dirs,logfiles,codes,logs,pws,pots,pos,kpts,params =[],[],[],[],[],[],[],[],[]

    # Main Program
    #-------------
    for d in files - filt:
        dir = rootpath +'/'+d
        ds = listdir(dir)

        assert 'runtime.json' in ds,'Failed sanity check'

        pwinps = glob(dir+'*/pw.inp')

        # DFTcode-specific stuff
        #------------------------
        if len(pwinps) == 1:
            code    = 'quantumespresso'
            logfile = dir+'/log'
            with open(pwinps[0],'r') as f:
                pwinp = f.read()

            poscar,potcar,kptcar = '','',''

        elif 'OUTCAR' in ds:
            code = 'vasp'
            logfile = dir+'/OUTCAR'
            pwinp = ''
            with open(dir+'/POSCAR','r') as f:
                poscar = f.read()
            with open(dir+'/POTCAR','r') as f:
                potcar = f.read()
            with open(dir+'/KPOINTS','r') as f:
                kptcar = f.read()

        else:
            code    = 'gpaw'
            logfile = dir+'/log'
            pwinp,poscar,potcar,kptcar = '','','',''

        # Common attributes
        #------------------
        with open(logfile,'r') as f:
            log = f.read()

        with open(dir+'/params.json','r') as f:
            paramjson = f.read()


        # Store results
        #------------
        dirs.append(dir)
        logfiles.append(logfile)
        codes.append(code)
        logs.append(log)
        pws.append(pwinp)
        pots.append(potcar)
        pos.append(poscar)
        kpts.append(kptcar)
        params.append(paramjson)

    return dirs,logfiles,codes,logs,pws,pots,pos,kpts,params

if __name__ == '__main__':
    import sys
    print(get_catalog_sherlock(sys.argv[1],''))
